"""Top-level package for bosonic-qiskit."""

# flake8: noqa

from c2qa.circuit import CVCircuit
from c2qa.qumoderegister import QumodeRegister

import c2qa.animate
import c2qa.discretize
import c2qa.kraus
import c2qa.operators
import c2qa.parameterized_unitary_gate
import c2qa.util
import c2qa.wigner



import qiskit
import qiskit_aer


def test_qiskit():
    """Verify we can do a simple QisKit circuit without our custom gates."""
    qr = qiskit.QuantumRegister(6)
    cr = qiskit.ClassicalRegister(6)
    circuit = qiskit.circuit.QuantumCircuit(qr, cr)
    circuit.cx(qr[0:1], qr[2])
    circuit.save_statevector()

    backend = qiskit_aer.AerSimulator()
    job = backend.run(circuit)
    result = job.result()
    state = result.get_statevector(circuit)

    assert result.success
    print(state)


def test_initialize(capsys):
    with capsys.disabled():
        # Successful with Qiskit v0.34.2, raises error with v0.35+
        qr = qiskit.QuantumRegister(1)
        circuit = qiskit.circuit.QuantumCircuit(qr)
        circuit.initialize([0, 1], qr[0])



import c2qa


def test_qumode_iterate(capsys):
    with capsys.disabled():
        num_qubits_per_qumode = 2
        num_qumodes = 3

        register = c2qa.QumodeRegister(
            num_qubits_per_qumode=num_qubits_per_qumode, num_qumodes=num_qumodes
        )
        index = 0

        for qumode in register:
            print(qumode)
            index += 1
            assert len(qumode) == num_qubits_per_qumode

        assert index == num_qumodes



from pathlib import Path


import c2qa
import numpy
import qiskit


def test_plot_zero(capsys):
    with capsys.disabled():
        qmr = c2qa.QumodeRegister(num_qumodes=1, num_qubits_per_qumode=4)
        qr = qiskit.QuantumRegister(size=1)
        circuit = c2qa.CVCircuit(qmr, qr)

        # dist = numpy.sqrt(numpy.pi) / numpy.sqrt(2)
        dist = 1.0

        # qr[0] will init to zero
        circuit.cv_initialize(0, qmr[0])

        circuit.h(qr[0])
        circuit.cv_c_d(dist, qmr[0], qr[0])

        state, result, fock_counts = c2qa.util.simulate(circuit)
        trace = c2qa.util.trace_out_qubits(circuit, state)
        c2qa.wigner.plot_wigner(circuit, trace, file="tests/zero.png", trace=False)


def test_plot_one(capsys):
    with capsys.disabled():
        qmr = c2qa.QumodeRegister(num_qumodes=1, num_qubits_per_qumode=2)
        qr = qiskit.QuantumRegister(size=1)
        cr = qiskit.ClassicalRegister(size=1)
        circuit = c2qa.CVCircuit(qmr, qr, cr)

        # qr[0] and cr[0] will init to zero
        circuit.cv_initialize(1, qmr[0])

        state, result, fock_counts = c2qa.util.simulate(circuit)
        # print("Qumode initialized to one:")
        # print(state)
        c2qa.wigner.plot_wigner(circuit, state, file="tests/one.png")


def test_plot_wigner_projection(capsys):
    with capsys.disabled():
        qmr = c2qa.QumodeRegister(num_qumodes=1, num_qubits_per_qumode=4)
        qr = qiskit.QuantumRegister(size=1)
        circuit = c2qa.CVCircuit(qmr, qr)

        # dist = numpy.sqrt(numpy.pi) / numpy.sqrt(2)
        dist = 1.0

        # qr[0] will init to zero
        circuit.cv_initialize(0, qmr[0])

        circuit.h(qr[0])
        circuit.cv_c_d(dist, qmr[0], qr[0])
        # circuit.cv_d(dist, qmr[0])

        c2qa.wigner.plot_wigner_projection(
            circuit, qr[0], file="tests/interference.png"
        )


def test_cat_state_wigner_plot(capsys):
    with capsys.disabled():
        num_qubits_per_qumode = 4
        dist = 2

        qmr = c2qa.QumodeRegister(
            num_qumodes=1, num_qubits_per_qumode=num_qubits_per_qumode
        )
        qr = qiskit.QuantumRegister(size=1)
        cr = qiskit.ClassicalRegister(size=1)
        circuit = c2qa.CVCircuit(qmr, qr, cr)

        circuit.initialize([1, 0], qr[0])
        circuit.cv_initialize(0, qmr[0])

        circuit.h(qr[0])
        circuit.cv_c_d(dist, qmr[0], qr[0])
        circuit.h(qr[0])
        circuit.measure(qr[0], cr[0])

        # conditional_state_vector=True will return two state vectors, one for 0 and 1 classical register value
        state, result, fock_counts = c2qa.util.simulate(
            circuit, conditional_state_vector=True
        )
        even_state = state["0x0"]
        odd_state = state["0x1"]

        wigner_filename = "tests/cat_wigner_even.png"
        c2qa.wigner.plot_wigner(
            circuit,
            even_state,
            file=wigner_filename,
            trace=True,
            axes_min=-6,
            axes_max=6,
        )
        assert Path(wigner_filename).is_file()

        wigner_filename = "tests/cat_wigner_odd.png"
        c2qa.wigner.plot_wigner(
            circuit,
            odd_state,
            file=wigner_filename,
            trace=True,
            axes_min=-6,
            axes_max=6,
        )
        assert Path(wigner_filename).is_file()

        # # Need to recreate circuit state prior to measure collapsing qubit state for projections
        # qmr = c2qa.QumodeRegister(num_qumodes=1, num_qubits_per_qumode=num_qubits_per_qumode)
        # qr = qiskit.QuantumRegister(size=1)
        # cr = qiskit.ClassicalRegister(size=1)
        # circuit = c2qa.CVCircuit(qmr, qr, cr)
        # circuit.initialize(state)

        # wigner_filename = "tests/repeat_projection_wigner.png"
        # c2qa.wigner.plot_wigner_projection(circuit, qr[0], file=wigner_filename)
        # assert Path(wigner_filename).is_file()


def test_wigner_mle(capsys):
    with capsys.disabled():
        num_qubits_per_qumode = 4
        dist = 2

        qmr = c2qa.QumodeRegister(
            num_qumodes=1, num_qubits_per_qumode=num_qubits_per_qumode
        )
        qr = qiskit.QuantumRegister(size=1)
        cr = qiskit.ClassicalRegister(size=1)
        circuit = c2qa.CVCircuit(qmr, qr, cr)

        circuit.initialize([1, 0], qr[0])
        circuit.cv_initialize(0, qmr[0])

        circuit.h(qr[0])
        circuit.cv_c_d(dist, qmr[0], qr[0])
        circuit.h(qr[0])

        state, result, fock_counts = c2qa.util.simulate(
            circuit, per_shot_state_vector=True
        )
        wigner = c2qa.wigner.wigner_mle(state)
        assert wigner is not None
        print(wigner)


def test_plot_wigner_snapshot(capsys):
    with capsys.disabled():
        num_qubits_per_qumode = 4
        dist = 2

        qmr = c2qa.QumodeRegister(
            num_qumodes=1, num_qubits_per_qumode=num_qubits_per_qumode
        )
        qr = qiskit.QuantumRegister(size=1)
        cr = qiskit.ClassicalRegister(size=1)
        circuit = c2qa.CVCircuit(qmr, qr, cr)

        circuit.initialize([1, 0], qr[0])
        circuit.cv_initialize(0, qmr[0])

        circuit.cv_snapshot()

        circuit.h(qr[0])

        circuit.cv_snapshot()
        circuit.cv_c_d(dist, qmr[0], qr[0])

        circuit.cv_snapshot()

        circuit.h(qr[0])

        circuit.cv_snapshot()

        state, result, fock_counts = c2qa.util.simulate(circuit)

        c2qa.wigner.plot_wigner_snapshot(circuit, result, "tests")


def test_plot_zero_contour(capsys):
    with capsys.disabled():
        data = numpy.zeros((200, 200)).tolist()
        filename = "tests/test_plot_zero_contour.png"
        c2qa.wigner.plot(data, file=filename)
        assert Path(filename).is_file()



import numpy
import qiskit
import qiskit_aer
import scipy.linalg


import c2qa


# Define parameters
num_qubits_per_qumode = 2
cutoff = 2**num_qubits_per_qumode
alpha = numpy.sqrt(numpy.pi)


def displacement_operator(arg):
    """Create displacement operator matrix"""
    a = numpy.diag(numpy.sqrt(range(1, cutoff)), k=1)
    a_dag = a.conj().T
    return scipy.linalg.expm((arg * a_dag) - (numpy.conjugate(arg) * a))


def displacemnt_gate(circuit, arg, qumode):
    circuit.unitary(displacement_operator(arg), qumode)


def conditional_displacement_gate(circuit, arg_0, arg_1, qbit, qumode):
    """Append a conditional displacement to the circuit
    Displace by arg_0 if qbit is 0, by arg_1 if qbit is 1."""

    op_0 = displacement_operator(arg_0)
    op_1 = displacement_operator(arg_1)

    circuit.append(
        qiskit.circuit.library.UnitaryGate(op_0).control(
            num_ctrl_qubits=1, ctrl_state=0
        ),
        [qbit] + qumode,
    )
    circuit.append(
        qiskit.circuit.library.UnitaryGate(op_1).control(
            num_ctrl_qubits=1, ctrl_state=1
        ),
        [qbit] + qumode,
    )


def qumode_initialize(circuit, fock_state, qumode):
    """Initialize the qumode to a Fock state."""

    value = numpy.zeros((cutoff,))
    value[fock_state] = 1

    circuit.initialize(value, qumode)


def run_displacement_calibration(enable_measure):
    """
    Run the simulation that has different state vector results on Windows vs Linux.

      - Create the qumode register as a QuantumRegiser where n qubits represent a qumode.
      - Create a QuantumRegister(1) to represent the control qubit.
      - Initialize the qubits.
      - Simulate the circuit.
    """

    # Instantiate QisKit registers & circuit
    qmr = qiskit.QuantumRegister(num_qubits_per_qumode)  # qumode register
    qr = qiskit.QuantumRegister(1)
    cr = qiskit.ClassicalRegister(1)
    circuit = qiskit.QuantumCircuit(qmr, qr, cr)

    # Initialize the qumode Fock state
    # qr[0] and cr[0] will init to zero
    qumode_initialize(circuit, 0, qmr[0:])

    circuit.h(qr[0])
    conditional_displacement_gate(circuit, alpha, -alpha, qr[0], qmr[0:])
    displacemnt_gate(circuit, 1j * alpha, qmr[0:])
    conditional_displacement_gate(circuit, -alpha, alpha, qr[0], qmr[0:])
    displacemnt_gate(circuit, -1j * alpha, qmr[0:])
    circuit.h(qr[0])

    circuit.save_statevector()

    if enable_measure:
        circuit.measure(qr[0], cr[0])

    backend = qiskit_aer.AerSimulator()
    circuit = qiskit.transpile(circuit, backend)
    job = backend.run(circuit)
    result = job.result()
    state = result.get_statevector(circuit)
    counts = result.get_counts(circuit)

    assert state.dim > 0
    assert counts

    # print(state)
    # print(counts.int_outcomes())


def test_displacement_calibration(capsys):
    with capsys.disabled():
        run_displacement_calibration(False)
        run_displacement_calibration(True)



import warnings


import qiskit
from qiskit import QuantumCircuit, QuantumRegister
from qiskit.circuit import Gate
from qiskit.circuit.library import UnitaryGate
from qiskit.circuit.parameter import ParameterExpression


class ParameterizedUnitaryGate(Gate):
    """UnitaryGate sublcass that stores the operator matrix for later reference by animation utility."""

    def __init__(
        self,
        op_func,
        params,
        num_qubits,
        cutoffs,
        label=None,
        duration=100,
        unit="ns",
        discretized_param_indices: list = [],
    ):
        """Initialize ParameterizedUnitaryGate

        Args:
            op_func (function): function to build operator matrix
            params (List): List of parameters to pass to op_func to build operator matrix (supports instances of Qiskit Parameter to be bound later)
            num_qubits (int): Number of qubits in the operator -- this would likely equate to (num_qubits_per_qumode * num_qumodes + num_ancilla).
            label (string, optional): Gate name. Defaults to None.
            duration (int, optional): Duration of gate used for noise modeling. Defaults to 100.
            unit (string, optional): Unit of duration (only supports those allowed by Qiskit).
            discretized_param_indices (list): list of int indices into self.params for parameters to be discretized. An empty list will discretize all params.
        """
        super().__init__(name=label, num_qubits=num_qubits, params=params, label=label)

        self.op_func = op_func

        self._parameterized = any(
            isinstance(param, ParameterExpression) and param.parameters
            for param in params
        )

        self.duration = duration
        self.unit = unit
        self.discretized_param_indices = discretized_param_indices
        self.cutoffs = cutoffs

    def __array__(self, dtype=None):
        """Call the operator function to build the array using the bound parameter values."""
        # return self.op_func(*map(complex, self.params)).toarray()
        values = []

        # Add parameters for op_func call
        for param in self.params:
            if isinstance(param, ParameterExpression):
                # if param.is_real():
                #     values.append(float(param))
                # else:
                #     values.append(complex(param))
                values.append(
                    complex(param)
                )  # just cast everything to complex to avoid errors in Ubuntu/MacOS vs Windows
            else:
                values.append(param)

        # Add cutoff for each parameter
        values.extend(self.cutoffs)

        # Conver array to tupple
        values = tuple(values)

        return self.op_func(*values).toarray()

    def _define(self):
        try:
            mat = self.to_matrix()
            q = QuantumRegister(self.num_qubits)
            qc = QuantumCircuit(q, name=self.name)
            rules = [
                (UnitaryGate(mat, self.label), [i for i in q], []),
            ]
            for instr, qargs, cargs in rules:
                qc._append(instr, qargs, cargs)

            self.definition = qc
        except:
            # warnings.warn("Unable to define gate, setting definition to None to prevent serialization errors for parameterized unitary gates.")
            self.definition = None

    def validate_parameter(self, parameter):
        """Gate parameters should be int, float, or ParameterExpression"""
        if isinstance(parameter, complex) or (
            isinstance(parameter, ParameterExpression) and not parameter.is_real()
        ):
            return parameter
        elif isinstance(parameter, str):  # accept strings as-is
            return parameter
        elif isinstance(parameter, list):
            return parameter
        else:
            return super().validate_parameter(parameter)

    def calculate_matrix(
        self, current_step: int = 1, total_steps: int = 1, keep_state: bool = False
    ):
        """Calculate the operator matrix by executing the selected function.
        Increment the parameters based upon the current and total steps.

        Args:
            current_step (int, optional): Current step within total_steps. Defaults to 1.
            total_steps (int, optional): Total steps to increment parameters. Defaults to 1.

        Returns:
            ndarray: operator matrix
        """
        if self.is_parameterized():
            raise NotImplementedError(
                "Unable to calculate incremental operator matrices for parameterized gate"
            )

        values = self.calculate_segment_params(current_step, total_steps, keep_state)

        # if self.inverse:
        #     result = scipy.sparse.linalg.inv(self.op_func(*values))
        # else:
        #     result = self.op_func(*values)
        result = self.op_func(*values)

        if hasattr(result, "toarray"):
            result = result.toarray()

        return result


def __calculate_segment_params(
    self, current_step: int = 1, total_steps: int = 1, keep_state: bool = False
):
    """
    Calculate the parameters at the current step. Return a tuples of the values.

     Args:
        current_step (int): 0-based current step index of the discretization
        total_steps (int): total number of discretization steps
        keep_state (bool): true if the state should be kept between discretization steps (i.e., if the discretization value should be 1/total_steps vs current_step/total_steps)

    Returns:
        discretized parameter values as tuple
    """
    if keep_state:
        param_fraction = 1 / total_steps
    else:
        param_fraction = current_step / total_steps

    values = []
    for index, param in enumerate(self.params):
        if (
            not hasattr(self, "discretized_param_indices")
            or len(self.discretized_param_indices) == 0
            or index in self.discretized_param_indices
        ):
            values.append(param * param_fraction)
        else:
            values.append(param)

    return tuple(values)


def __calculate_segment_duration(
    self, current_step: int = 1, total_steps: int = 1, keep_state: bool = False
):
    """Calculate the duration at the current step. Return a tuple of the (duration, unit)."""
    frame_duration = None

    if self.duration:
        if keep_state:
            fraction = 1 / total_steps
        else:
            fraction = current_step / total_steps

        frame_duration = self.duration * fraction

    return frame_duration, self.unit


# Monkey patch Qiskit Instruction to support animating base Qiskit Instruction
qiskit.circuit.instruction.Instruction.calculate_segment_params = (
    __calculate_segment_params
)
qiskit.circuit.instruction.Instruction.calculate_segment_duration = (
    __calculate_segment_duration
)



from pathlib import Path


import c2qa
import numpy
import qiskit


def assert_changed(state, result):
    assert result.success
    # print()
    # print(circuit.draw("text"))
    # print(state)

    # TODO - better understand what the state vector results should be
    assert count_nonzero(state) > 1


def count_nonzero(statevector: qiskit.quantum_info.Statevector):
    """Re-implement numpy.count_nonzero using numpy.isclose()."""
    nonzero = len(statevector.data)
    for state in statevector.data:
        if numpy.isclose(state, 0):
            nonzero -= 1

    return nonzero


def create_conditional(num_qumodes: int = 2, num_qubits_per_qumode: int = 2):
    qmr = c2qa.QumodeRegister(num_qumodes, num_qubits_per_qumode)
    qr = qiskit.QuantumRegister(2)
    circuit = c2qa.CVCircuit(qmr, qr)

    for qumode in range(num_qumodes):
        circuit.cv_initialize(0, qmr[qumode])

    circuit.initialize([0, 1], qr[1])  # qr[0] will init to zero

    return circuit, qmr, qr


def test_parameterized_displacement(capsys):
    with capsys.disabled():
        circuit, qmr, qr = create_conditional()

        alpha = qiskit.circuit.Parameter("alpha")
        circuit.cv_d(alpha, qmr[0])

        bound_circuit = circuit.assign_parameters({alpha: 3.14})

        state, result, fock_counts = c2qa.util.simulate(bound_circuit)
        assert_changed(state, result)


def test_complex_literals(capsys):
    with capsys.disabled():
        # a = qiskit.circuit.Parameter('𝛼')

        qmr = c2qa.QumodeRegister(1, num_qubits_per_qumode=4)
        qbr = qiskit.QuantumRegister(1)

        minimal_circuit = c2qa.CVCircuit(qmr, qbr)

        minimal_circuit.h(qbr[0])

        minimal_circuit.cv_c_d(1j * 1, qmr[0], qbr[0])

        # bound_circuit = minimal_circuit.assign_parameters({a: 1})

        c2qa.util.simulate(minimal_circuit)


def test_complex_parameters(capsys):
    with capsys.disabled():
        a = qiskit.circuit.Parameter("𝛼")

        qmr = c2qa.QumodeRegister(1, num_qubits_per_qumode=4)
        qbr = qiskit.QuantumRegister(1)

        minimal_circuit = c2qa.CVCircuit(qmr, qbr)

        minimal_circuit.h(qbr[0])

        minimal_circuit.cv_c_d(1j * a, qmr[0], qbr[0])

        bound_circuit = minimal_circuit.assign_parameters({a: 1})
        c2qa.util.simulate(bound_circuit)


def test_complex_parameters_float(capsys):
    with capsys.disabled():
        a = qiskit.circuit.Parameter("𝛼")

        qmr = c2qa.QumodeRegister(1, num_qubits_per_qumode=4)
        qbr = qiskit.QuantumRegister(1)

        minimal_circuit = c2qa.CVCircuit(qmr, qbr)

        minimal_circuit.h(qbr[0])

        minimal_circuit.cv_c_d(1j * a, qmr[0], qbr[0])

        bound_circuit = minimal_circuit.assign_parameters({a: 2})
        c2qa.util.simulate(bound_circuit)



import math
import warnings

import numpy as np
import qiskit
import qiskit.quantum_info
from qiskit.quantum_info import Statevector, DensityMatrix
import qiskit_aer


from c2qa import CVCircuit
from c2qa.discretize import discretize_circuits


def flatten(l):
    return [item for sublist in l for item in sublist]


def cv_ancilla_fock_measure(circuit, list_qumodes_to_sample: list, qmr_number: int = 0):
    """Simulate a circuit with an appended binary search for boson number, and determine the Fock state of a set of qumodes using
    phase kickback on the qubit. For more information, see Curtis et al., PRA (2021) and Wang et al., PRX (2020).

    Returns the Fock state of the qumodes in list_qumodes_to_sample, in qumode register qmr_number.
    """
    # Count number of qubits in circuit so far
    num_qubits = len(flatten(circuit._qubit_regs))
    # Collect qumode register from circuit
    qmr = circuit.qmregs[qmr_number]
    # Add one ancilla qubits to the circuit per qumode to measure
    qbr_extra = qiskit.QuantumRegister(
        size=len(list_qumodes_to_sample), name="qbr_sampling"
    )
    # Add classical bits to readout measurement results
    cbr_extra = qiskit.ClassicalRegister(
        len(list_qumodes_to_sample) * circuit.num_qubits_per_qumode, name="cbr_sampling"
    )
    circuit.add_register(qbr_extra, cbr_extra)
    # Iterate over the qumodes
    qumode_counter = 0
    for j in list_qumodes_to_sample:
        # Set the initial maximum Fock state
        max = circuit.cutoff
        # Iterate a number of time corresponding to the number of bits required to represent the maximum Fock state in binary (remove useless characters at the front)
        for iteration in range(len(bin(circuit.cutoff)) - 3):
            # Make sure the ancilla qubit is always reset to 0
            circuit.initialize("0", qbr_extra[qumode_counter])
            # Apply a circuit which flips the ancilla if the qumode occupation is odd etc. see (Curtis et al., PRA, 2021 and Wang et al.,  PRX, 2020)
            circuit.cv_c_pnr(
                max, qmr[list_qumodes_to_sample[j]], qbr_extra[qumode_counter]
            )
            # Measure the qubit onto the classical bits (from left to right)
            classical_bit = (
                circuit.num_qubits_per_qumode
                - 1
                - iteration
                + (qumode_counter * circuit.num_qubits_per_qumode)
            )
            circuit.measure(qbr_extra[qumode_counter], classical_bit)
            # Update the maximum value for the SNAP gate creation
            max = int(max / 2)
        circuit.barrier()
        qumode_counter += 1
    # Simulate circuit with a single shot
    _, result, _ = simulate(circuit, shots=1)
    # Return integer value of boson number occupation, converted from the bits which make up a binary number
    print(result.get_counts())
    full_set_of_binary = list(result.get_counts().keys())[0].encode("ascii")
    results_integers = np.zeros([len(list_qumodes_to_sample)])
    for j in range(len(list_qumodes_to_sample)):
        binary_number = full_set_of_binary[
            j
            * circuit.num_qubits_per_qumode : ((j + 1) * circuit.num_qubits_per_qumode)
        ]
        print(binary_number)
        results_integers[-(j + 1)] = int(binary_number, 2)
    return results_integers


def stateread(
    stateop, numberofqubits, numberofmodes, cutoff, verbose=True, little_endian=False
):
    """Print values for states of qubits and qumodes using the result of a
    simulation of the statevector, e.g. using stateop, _, _, _ = c2qa.util.simulate(circuit).

    Returns the states of the qubits and the Fock states of the qumodes with respective amplitudes.
    """
    st = np.array(stateop)  # convert state to np.array
    amp_cv = []
    amp_qb = []
    state = []

    cutoff = 2 ** int(
        np.ceil(np.log2(cutoff))
    )  # The cutoff needs to be a power of 2 for this code to work

    for i in range(len(st)):
        res = st[
            i
        ]  # go through vector element by element and find the positions of the non-zero elements with next if clause
        if np.abs(res) > 1e-10:
            pos = i  # position of amplitude (non-zero real)
            # print("position of non-zero real amplitude: ", pos, " res = ", res)
            sln = len(st)  # length of the state vector

            # Find the qubit states
            qbst = np.empty(numberofqubits, dtype="int")  # stores the qubit state
            iqb = 0  # counts up until the total number of qubits is reached

            # which half of the vector the amplitude is in is the state of the
            # first qubit because of how the kronecker product is made
            while iqb < numberofqubits:
                # if the amplitude is in the first half of the state vector or remaining statevector
                if pos < sln / 2:
                    qbst[iqb] = int(0)  # then the qubit is in 0
                else:
                    # if the amplitude is in the second half then it is in 1
                    qbst[iqb] = int(1)

                    # if the amplitude is in the second half of the statevector,
                    # then to find out the state of the other qubits and cavities
                    # then we remove the first half of the statevector for simplicity
                    # because it corresponds to the qubit being in 0 which isn't the case.
                    pos = pos - (sln / 2)
                    # print("pos (sln/2)", pos, "sln ",sln)

                # only consider the part of the statevector corresponding to the qubit state which has just been discovered
                sln = sln / 2

                # up the qubit counter to start finding out the state of the next qubit
                iqb = iqb + 1
            amp_qb.append((qbst * (np.abs(res) ** 2)).tolist())

            # Find the qumode states
            qmst = np.empty(
                numberofmodes, dtype="int"
            )  # will contain the Fock state of each mode
            # print("qmst starting in ", qmst)
            iqm = 0  # counts up the number of modes
            # print("position is now: ",pos)
            while iqm < numberofmodes:
                # print("mode counter iqm ", iqm)
                # print("cutoff ", cutoff)
                # print("length of vector left to search: sln ", sln)

                # length of a division is the length of the statevector divided
                # by the cutoff of the hilbert space (which corresponds to the
                # number of fock states which a mode can have)
                lendiv = sln / cutoff
                # print("lendiv (sln/cutoff)", lendiv)
                val = pos / lendiv
                # print("rough estimate of the position of the non-zero element: val (pos/lendiv) ", val)
                fock = math.floor(val)
                # print("Fock st resulting position in Kronecker product (math.floor(val)) ", fock)
                qmst[iqm] = fock

                # remove a number of divisions to then search a subsection of the Kronecker product
                pos = pos - (fock * lendiv)
                # print("new position for next order of depth of Kronecker product/pos: (pos-(fock*lendiv)) ",pos)

                # New length of vector left to search
                sln = sln - ((cutoff - 1) * lendiv)
                # print("New length of vector left to search: sln (sln-((cutoff-1)*lendiv))", sln)
                iqm = iqm + 1
            amp_cv.append((qmst * (np.abs(res) ** 2)).tolist())

            state.append([qmst.tolist(), qbst.tolist(), res])

            if verbose:
                if little_endian:
                    qmstr = ["".join(str(item)) for item in qmst]
                    qbstr = ["".join(str(item)) for item in qbst]
                    print(
                        "qumodes: ",
                        "".join(qmstr),
                        " qubits: ",
                        "".join(qbstr),
                        "    with amplitude: {0:.3f} {1} i{2:.3f}".format(
                            res.real, "-" if res.imag < 0 else "+", abs(res.imag)
                        ),
                        "(little endian)",
                    )
                else:
                    qmstr = ["".join(str(item)) for item in qmst[::-1]]
                    qbstr = ["".join(str(item)) for item in qbst[::-1]]
                    print(
                        "qumodes: ",
                        "".join(qmstr),
                        " qubits: ",
                        "".join(qbstr),
                        "    with amplitude: {0:.3f} {1} i{2:.3f}".format(
                            res.real, "-" if res.imag < 0 else "+", abs(res.imag)
                        ),
                        "(big endian)",
                    )

    occupation_cv = [sum(i) for i in zip(*amp_cv)]
    occupation_qb = [sum(i) for i in zip(*amp_qb)]

    if not little_endian:
        for i in range(len(state)):
            state[i][0] = state[i][0][::-1]
            state[i][1] = state[i][1][::-1]
        occupation_cv = occupation_cv[::-1]
        occupation_qb = occupation_qb[::-1]

    return [occupation_cv, occupation_qb], state


def counts_to_fockcounts(
    circuit: CVCircuit, result: qiskit.result.result.Result, counts: dict = None
):
    """Convert counts dictionary from Fock-basis binary representation into
    base-10 Fock basis (qubit measurements are left unchanged). Accepts the object returned by
    jobs.result(), along with the entire circuit.

    Args:
        result: dict() of results, as returned by job.result(), for a circuit which used cv_measure()
        circuit: CVCircuit

    Returns:
        A new counts dict() which lists measurement results for the qubits and
        qumodes in circuit in little endian order, with Fock-basis
        qumode measurements reported as a base-10 integer.
    """
    if not counts:
        counts = result.get_counts()
    qumode_bit_mapping = _final_qumode_mapping(circuit)

    newcounts = {}
    for key in counts:
        max_iter_index = len(key) - 1
        newkey = key

        # Using the nested list of qumode bit mappings, convert the relevant bits to base-10 integer and
        # form new key by concatenating the unchanged bits of key around the decimal value.
        for index in range(len(key)):
            for qumode in qumode_bit_mapping:
                if index == min(qumode):
                    fock_decimal = str(
                        int(
                            key[
                                max_iter_index
                                - max(qumode) : max_iter_index
                                - min(qumode)
                                + 1
                            ],
                            base=2,
                        )
                    )
                    newkey = (
                        newkey[: max_iter_index - max(qumode)]
                        + fock_decimal
                        + newkey[max_iter_index - min(qumode) + 1 :]
                    )

        newcounts[newkey] = counts[key]

    return newcounts


def _final_qumode_mapping(circuit):
    """
    Return the classical bits that active qumode qubits are mapped onto. Bits corresponding to distinct qumodes are grouped together
    """
    final_measurement_mapping = _final_measurement_mapping(circuit)
    active_qumode_bit_indices_grouped = []

    # For each qumode qubit group, extract list of bits that map to qubits in group. Append list only if list is not empty
    for qumode_qubit_group in circuit.qumode_qubits_indices_grouped:
        qumode_bit_group = [
            key
            for key, val in final_measurement_mapping.items()
            for qubit in qumode_qubit_group
            if val == qubit
        ]

        if qumode_bit_group != []:
            active_qumode_bit_indices_grouped.append(qumode_bit_group)

    # Sort nested list by first item in each sublist
    active_qumode_bit_indices_grouped = sorted(
        active_qumode_bit_indices_grouped, key=lambda l: l[0]
    )

    return active_qumode_bit_indices_grouped


# This code is part of Mthree.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=no-name-in-module
def _final_measurement_mapping(circuit):
    """Return the measurement mapping for the circuit.

    Dict keys label classical bits, whereas the values indicate the
    physical qubits that are measured to produce those bit values.

    Parameters:
        circuit (QuantumCircuit): Input Qiskit QuantumCircuit.

    Returns:
        dict: Mapping of classical bits to qubits for final measurements.
    """
    active_qubits = list(range(circuit.num_qubits))
    active_cbits = list(range(circuit.num_clbits))

    # Map registers to ints
    qint_map = {}
    for idx, qq in enumerate(circuit.qubits):
        qint_map[qq] = idx

    cint_map = {}
    for idx, qq in enumerate(circuit.clbits):
        cint_map[qq] = idx

    # Find final measurements starting in back
    qmap = []
    cmap = []
    for item in circuit._data[::-1]:
        if item[0].name == "measure":
            cbit = cint_map[item[2][0]]
            qbit = qint_map[item[1][0]]
            if cbit in active_cbits and qbit in active_qubits:
                qmap.append(qbit)
                cmap.append(cbit)
                active_cbits.remove(cbit)

        if not active_cbits or not active_qubits:
            break
    mapping = {}
    if cmap and qmap:
        for idx, qubit in enumerate(qmap):
            mapping[cmap[idx]] = qubit

    # Sort so that classical bits are in numeric order low->high.
    mapping = dict(sorted(mapping.items(), key=lambda item: item[0]))
    return mapping


def measure_all_xyz(circuit: qiskit.QuantumCircuit):
    """Use QuantumCircuit.measure_all() to measure all qubits in the X, Y, and Z basis.

    Returns state, result, fockcounts tuples each for the X, Y, and Z basis.

    Args:
        circuit (qiskit.QuantumCircuit): circuit to measure qubits one

    Returns:
        x,y,z state & result tuples: (state, result, fockcounts) tuples for each x,y,z measurements
    """

    # QuantumCircuit.measure_all(False) returns a copy of the circuit with measurement gates.
    circuit_z = circuit.measure_all(False)
    state_z, result_z, fockcounts_z = simulate(circuit_z)

    circuit_x = circuit.copy()
    for qubit in circuit_x.qubits:
        circuit_x.h(qubit)
    circuit_x.measure_all()  # Add measure gates in-place
    state_x, result_x, fockcounts_x = simulate(circuit_x)

    circuit_y = circuit.copy()
    for qubit in circuit_y.qubits:
        circuit_y.sdg(qubit)
        circuit_y.h(qubit)
    circuit_y.measure_all()  # Add measure gates in-place
    state_y, result_y, fockcounts_y = simulate(circuit_y)

    return (
        (state_x, result_x, fockcounts_x),
        (state_y, result_y, fockcounts_y),
        (state_z, result_z, fockcounts_z),
    )


def get_probabilities(result: qiskit.result.Result):
    """Calculate the probabilities for each of the result's counts.

    Args:
        result (qiskit.result.Result): QisKit result to calculate probabilities from

    Returns:
        dict: probablity dictionary of each state
    """
    shots = 0
    counts = result.get_counts()
    for count in counts:
        shots += counts[count]
    probs = {}
    for count in counts:
        probs[count] = counts[count] / shots

    return probs


def simulate(
    cvcircuit: CVCircuit,
    shots: int = 1024,
    return_fockcounts: bool = True,
    add_save_statevector: bool = True,
    conditional_state_vector: bool = False,
    per_shot_state_vector: bool = False,
    noise_model=None,
    noise_passes=None,
    max_parallel_threads: int = 0,
    discretize: bool = False,
):
    """Convenience function to simulate using the given backend.

    Handles calling into QisKit to simulate circuit using defined simulator.

    Args:
        circuit (CVCircuit): circuit to simulate
        shots (int, optional): Number of simulation shots. Defaults to 1024.
        return_fockcounts (bool, optional): Set to True if measurement results should be returned. Defaults to False
        add_save_statevector (bool, optional): Set to True if a state_vector instruction
                                               should be added to the end of the circuit. Defaults to True.
        conditional_state_vector (bool, optional): Set to True if the saved state vector should be contional
                                                   (each state value gets its own state vector). Defaults to False.
        per_shot_state_vector (bool, optional): Set to Ture if the simulation should return a separate state vector for
                                                every simulation shot.
        noise_model (NoiseModel, optional): Custom noise model to pass into AerSimulator to use on each simulation shot
        noise_passes (list of LocalNoisePass, optional): Custom noise pass to apply on each gate.
        max_parallel_threads (int, opational): Sets the maximum number of CPU cores used by OpenMP for parallelization.
                                               If set to 0 the maximum will be set to the number of CPU cores (Default: 0).
        discretize (bool, optional): Set to True if circuit should be discretized to apply noise passes. Defaults to False.

    Returns:
        tuple: (state, result, fock_counts) tuple from [optionally discretized] simulations
    """

    if discretize and not noise_passes:
        warnings.warn(
            "Discretization of circuit intended for use with noise passes, but none provided"
        )

    if discretize:
        sim_circuit = discretize_circuits(cvcircuit)[-1]
    else:
        sim_circuit = cvcircuit

    # If this is false, the user must have already called save_statevector!
    if add_save_statevector:
        sim_circuit.save_statevector(
            conditional=conditional_state_vector, pershot=per_shot_state_vector
        )

    # Run noise pass, if provided
    if noise_passes:
        if not isinstance(noise_passes, list):
            noise_passes = [noise_passes]

        for noise_pass in noise_passes:
            circuit_compiled = noise_pass(sim_circuit)
    else:
        circuit_compiled = sim_circuit

    # Transpile for simulator
    simulator = qiskit_aer.AerSimulator()
    circuit_compiled = qiskit.transpile(circuit_compiled, simulator)

    # Run and get statevector
    result = simulator.run(
        circuit_compiled,
        shots=shots,
        max_parallel_threads=max_parallel_threads,
        noise_model=noise_model,
    ).result()

    # The user may have added their own circuit.save_statevector
    state = None
    if len(result.results):
        try:
            if conditional_state_vector or per_shot_state_vector:
                # Will get a dictionary of state vectors, one for each classical register value
                state = result.data()["statevector"]
            else:
                state = Statevector(result.get_statevector(circuit_compiled))
        except Exception:
            state = None  # result.get_statevector() will fail if add_save_statevector is false

    if add_save_statevector:
        sim_circuit.data.pop()  # Clean up by popping off the SaveStatevector instruction

    if return_fockcounts and add_save_statevector:
        try:
            fockcounts = counts_to_fockcounts(sim_circuit, result, result.get_counts())
            return (state, result, fockcounts)
        except:
            raise Exception("counts_to_fockcounts() was not able to execute")
    else:
        return (state, result, None)


def _find_cavity_indices(circuit: CVCircuit):
    """
    Return the indices of the cavities from the circuit

    I.e., the indices to the qubits representing the bosonic modes.
    """

    # Find indices of qubits representing qumodes
    qmargs = []
    for reg in circuit.qmregs:
        qmargs.extend(reg.qreg)

    # Trace over the qubits representing qumodes
    index = 0
    indices = []
    for qubit in circuit.qubits:
        if qubit in qmargs:
            indices.append(index)
        index += 1

    return indices


def _find_qubit_indices(circuit: CVCircuit):
    """
    Return the indices of the qubits from the circuit that are not in a QumodeRegister

    I.e., the indices to the qubits themselves, not the qubits representing the bosonic modes.
    """

    # Find indices of qubits representing qumodes
    qmargs = []
    for reg in circuit.qmregs:
        qmargs.extend(reg.qreg)

    # Trace over the qubits not representing qumodes
    index = 0
    indices = []
    for qubit in circuit.qubits:
        if qubit not in qmargs:
            indices.append(index)
        index += 1

    return indices


def trace_out_qumodes(circuit: CVCircuit, state_vector):
    """Return reduced density matrix of the qubits by tracing out the cavities of the CVCircuit from the given Fock state vector.

    Args:
        circuit (CVCircuit): circuit yielding the results to trace over
        state_vector (Statevector): simulation results to trace over

    Returns:
        DensityMatrix: density matrix of the qubits from a partial trace over the cavities
    """

    indices = _find_cavity_indices(circuit)

    return qiskit.quantum_info.partial_trace(state_vector, indices)


def trace_out_qubits(circuit: CVCircuit, state_vector):
    """Return reduced density matrix of the cavities by tracing out the all qubits of the CVCircuit from the given Fock state vector.

    Args:
        circuit (CVCircuit): circuit with results to trace (to find Qubit index)
        state_vector (Statevector or DensityMatrix): simulation results to trace over

    Returns:
        DensityMatrix: partial trace
    """

    indices = _find_qubit_indices(circuit)

    return qiskit.quantum_info.partial_trace(state_vector, indices)


def cv_partial_trace(circuit: CVCircuit, state_vector, qubits: list):
    """Return reduced density matrix over the given Qiskit Qubits.

    First find the indices of the given Qubits, then call qiskit.quantum_info.partial_trace

    Args:
        circuit (CVCircuit): circuit with results to trace (to find Qubit index)
        state_vector (Statevector or DensityMatrix): simulation results to trace over
        qubits (list): list of Qiskit Qubit to trace over

    Returns:
        DensityMatrix: partial trace"""

    if not isinstance(qubits, list):
        qubits = [qubits]
    indices = circuit.get_qubit_indices(qubits)

    return qiskit.quantum_info.partial_trace(state_vector, indices)


def fockmap(matrix, fock_input, fock_output, amplitude=[]):
    """Generates matrix corresponding to some specified mapping of Fock states for a single qumode.
    First feed function empty matrix, then call fmap_matrix however many times needed to fully define intended mapping.
    Maps ith element in fock_input to ith element in fock_output with amplitude specified by ith element in amplitude.
    If amplitude is left blank, function assumes amp = 1 for all mappings.

    Two use cases
    1) int + list datatype combination (length of amp list must match length of either fock_input or fock_output, whichever is longer):
    >fockmap(matrix, 1, [0, 1])
    ->> ``|0><1| + |1><1|``

    >fockmap(matrix, [3, 2], 0, [0.5j, 1])
    ->> ``0.5j|0><3| + |0><2|``

    2) list datatype
    >fockmap(matrix, [3, 2], [2, 1], [0.1j, 0.8])
    ->> ``0.1j|2><3| + 0.8|1><2|``

    >fockmap(matrix, [1, 1], [2, 4])
    ->> ``|2><1| + |4><1|``


    Args:
        matrix (nested list/np.array): Matrix that you want to change
        fock_input (int/list): Input state(s) for mapping, corresponds to bra
        fock_output (int/list): Output states(s) for mapping, corresponds to ket
        amplitude (int/float/complex/list/ndarray): Amplitudes corresponding to final mapped states

    Returns:
        np.array: Edited matrix"""

    # Convert args to lists for convenience of computation
    if isinstance(fock_input, int):
        fock_input = [fock_input]
    elif isinstance(fock_input, (float, complex, str, bool)):
        raise TypeError("Please ensure that your fock_input value is an int")

    if isinstance(fock_output, int):
        fock_output = [fock_output]
    elif isinstance(fock_output, (float, complex, str, bool)):
        raise TypeError("Please ensure that your fock_output value is an int")

    if isinstance(amplitude, (int, float, complex)):
        amplitude = [amplitude]
    elif isinstance(amplitude, np.ndarray):
        amplitude = list(amplitude)

    # If user inputs python list instead of np array
    matrix = np.array(matrix, dtype=complex)

    # Default amplitude is 1 for all states, unless otherwise specified
    if not amplitude:  # Empty lists evaluate as False
        amplitude = [1 for i in range(max(len(fock_input), len(fock_output)))]

    ## Error flags
    # Length of amplitude list must match the greater of either input or output
    if not (len(amplitude) == max(len(fock_input), len(fock_output))):
        raise ValueError(
            "Please ensure that that length of amplitude arg matches length of either input or output list."
        )
    # Datatype of args check
    if not (
        isinstance(fock_input, list)
        & isinstance(fock_output, list)
        & isinstance(amplitude, (list, np.ndarray))
    ):
        raise TypeError(
            "Please ensure that datatypes of input and output states are either int or list."
        )
    # Check for cutoff
    n, m = matrix.shape
    if n != m:
        raise ValueError("Matrix given is not square")
    if any(i >= n for i in fock_input) or any(j >= n for j in fock_output):
        raise ValueError("Fock state(s) greater than cutoff.")

    ## Use cases
    # 1. Int + int datatype for input/output args.
    if (len(fock_input) == 1) & (len(fock_output) == 1):
        if len(amplitude) > 1:
            raise ValueError(
                "Please ensure that only a single amplitude value is provided, as there is only 1 mapping provided"
            )
        if matrix[fock_output[0], fock_input[0]] != 0:
            print(
                "Warning: Existing element for |{}><{}| will be overwritten".format(
                    fock_output[0], fock_input[0]
                )
            )

        matrix[fock_output[0], fock_input[0]] = amplitude[0]

        return matrix

    # 2. Int + list datatype for input/output args. Length of amplitude must match length of list
    elif (len(fock_input) == 1) & (len(fock_output) > 1):
        for i in range(len(fock_output)):
            if matrix[fock_output[i], fock_input[0]] != 0:
                print(
                    "Warning: Existing element for |{}><{}| will be overwritten".format(
                        fock_output[i], fock_input[0]
                    )
                )

            matrix[fock_output[i], fock_input[0]] = amplitude[i]

        return matrix

    elif (len(fock_output) == 1) & (len(fock_input) > 1):
        for i in range(len(fock_input)):
            if matrix[fock_output[0], fock_input[i]] != 0:
                print(
                    "Warning: Existing element for |{}><{}| will be overwritten".format(
                        fock_output[0], fock_input[i]
                    )
                )

            matrix[fock_output[0], fock_input[i]] = amplitude[i]

        return matrix

    # 3. List datatype input/output/amp args. Lengths of all must match
    elif len(fock_input) == len(fock_output) == len(amplitude):

        for i in range(len(fock_input)):
            if matrix[fock_output[i], fock_input[i]] != 0:
                print(
                    "Warning: Existing element for |{}><{}| will be overwritten".format(
                        fock_output[i], fock_input[i]
                    )
                )

            matrix[fock_output[i], fock_input[i]] = amplitude[i]

        return matrix

    else:
        raise ValueError("Please ensure that your args are correctly defined.")


def avg_photon_num(circuit: CVCircuit, state, decimals: int = 2):
    """Returns average photon number of state for each qumode within the circuit using the number operator.

    Args:
        circuit (CVCircuit): Circuit definine qumodes present in given state
        state (Statevector or DensityMatrix): full state to operate on
        decimals: Determines precision of calculation

    Returns:
        float: Average photon number to specified precision
    """

    qumode_qubits = circuit.qumode_qubits_indices_grouped
    averages = []
    for qumode in range(len(qumode_qubits)):
        traced_qubits = []
        for traced_qumode in range(len(qumode_qubits)):
            if traced_qumode != qumode:
                traced_qubits.extend(qumode_qubits[traced_qumode])
        traced_state = qiskit.quantum_info.partial_trace(state, traced_qubits)
        averages.append(qumode_avg_photon_num(traced_state, decimals))

    return averages


def qumode_avg_photon_num(state, decimals: int = 2):
    """Returns average photon number of an individual qumode's state using the number operator.

    Args:
        state (Statevector or DensityMatrix): State to operate on for an individual qumode
        decimals: Determines precision of calculation

    Returns:
        float: Average photon number to specified precision
    """

    # Generate number operator based on dimension of state
    dim = state.dim
    N = np.diag(range(dim))

    # Normalise state
    if isinstance(state, Statevector):
        for_norm = state.inner(state)
    elif isinstance(state, DensityMatrix):
        for_norm = state.trace()
    else:
        raise TypeError(
            "Only Statevector or DensityMatrix are accepted as valid types."
        )

    # Calculate average photon number
    avg_photon = state.expectation_value(N) / for_norm

    if round(avg_photon.imag, 6) != 0:
        raise Exception(
            "Magnitude of average photon is complex, check inputs. Imaginary portion = {}".format(
                avg_photon.imag
            )
        )

    return np.round(avg_photon.real, decimals)



import math
import multiprocessing
import os
import pathlib


import matplotlib
import matplotlib.animation
import matplotlib.pyplot as plt
import numpy
import qiskit


from c2qa.circuit import CVCircuit
from c2qa.discretize import discretize_circuits, discretize_single_circuit
from c2qa.wigner import simulate_wigner, simulate_wigner_multiple_statevectors


def animate_wigner(
    circuit: CVCircuit,
    qubit: qiskit.circuit.quantumcircuit.QubitSpecifier = None,
    cbit: qiskit.circuit.quantumcircuit.ClbitSpecifier = None,
    animation_segments: int = 10,
    discretize_epsilon: float = None,
    shots: int = 1,
    file: str = None,
    axes_min: int = -6,
    axes_max: int = 6,
    axes_steps: int = 200,
    processes: int = None,
    keep_state: bool = True,
    noise_passes=None,
    sequential_subcircuit: bool = False,
    draw_grid: bool = False,
    trace: bool = True,
    bitrate: int = -1,
):
    """Animate the Wigner function at each step defined in the given CVCirctuit.

    This assumes the CVCircuit was simulated with an animation_segments > 0 to
    act as the frames of the generated movie.

    The ffmpeg binary must be on your system PATH in order to execute this
    function. See https://ffmpeg.org/download.html to download and install on your system.

    Args:
        circuit (CVCircuit): circuit to simulate and plot
        qubit ([type]): Qubit to measure, if performing Hadamard measure for use with cat states. Defaults to None.
        cbit ([type]): Classical bit to measure into, if performing Hadamard measure for use with cat states. Defaults to None.
        animation_segments (int, optional): Number of segments to split each gate into for animation. Defaults to 10.
        shots (int, optional): Number of simulation shots per frame. Defaults to 1.
        file (str, optional): File path to save (supported formats include MP4 with ffmpeg installed, animated GIF, and APNG).
                              If None, return plot. Defaults to None.
        axes_min (int, optional): Minimum axes plot value. Defaults to -6.
        axes_max (int, optional): Maximum axes plot value. Defaults to 6.
        axes_steps (int, optional): Steps between axes ticks. Defaults to 200.
        processes (int, optional): Number of parallel Python processes to start.
                                   If None, perform serially in main process. Defaults to None.
        keep_state (bool, optional): True if each frame builds on the previous frame's state vector.
                                     False if each frame starts over from the beginning of the circuit.
                                     If True, it requires sequential simulation of each frame.
        noise_passes (list of Qiskit noise passes, optional): noise passes to apply
        sequential_subcircuit (bool, optional): boolean flag to animate subcircuits as one gate (False) or as sequential
                                                gates (True). Defautls to False.
        draw_grid (bool, optional): True if grid lines should be drawn on plot. Defaults to False.
        trace (bool, optional):  True if qubits should be tracedfor each frame prior to calculating Wigner function. Defaults to True.

    Returns:
        [type]: [description]
    """

    if qubit or cbit:
        w_fock, xvec = __discretize_wigner_with_measure(
            circuit,
            qubit,
            cbit,
            animation_segments,
            shots,
            axes_min,
            axes_max,
            axes_steps,
            processes,
            keep_state,
            noise_passes,
            sequential_subcircuit,
            trace,
        )
    else:
        w_fock, xvec = __discretize_wigner_without_measure(
            circuit,
            animation_segments,
            discretize_epsilon,
            shots,
            axes_min,
            axes_max,
            axes_steps,
            noise_passes,
            sequential_subcircuit,
            trace,
        )

    # Remove None values in w_fock if simulation didn't produce results
    w_fock = [i for i in w_fock if i is not None]

    # Animate w_fock Wigner function results
    # Create empty plot to animate
    fig, ax = plt.subplots(constrained_layout=True)

    # Animate
    anim = matplotlib.animation.FuncAnimation(
        fig=fig,
        init_func=_animate_init,
        func=_animate,
        frames=len(w_fock),
        fargs=(fig, ax, xvec, w_fock, file, draw_grid),
        interval=200,
        repeat=True,
    )

    # Save to file using ffmpeg, Pillow (GIF, APNG), or display
    if file:
        save_animation(anim, file, bitrate)

    return anim


def __discretize_wigner_with_measure(
    circuit: CVCircuit,
    qubit: qiskit.circuit.quantumcircuit.QubitSpecifier = None,
    cbit: qiskit.circuit.quantumcircuit.ClbitSpecifier = None,
    animation_segments: int = 10,
    shots: int = 1,
    axes_min: int = -6,
    axes_max: int = 6,
    axes_steps: int = 200,
    processes: int = None,
    keep_state: bool = True,
    noise_passes=None,
    sequential_subcircuit: bool = False,
    trace: bool = True,
):
    circuits = discretize_circuits(
        circuit, animation_segments, keep_state, qubit, cbit, sequential_subcircuit
    )

    # Calculate the Wigner functions for each frame
    if not processes or processes < 1:
        processes = math.floor(multiprocessing.cpu_count() / 2)
        processes = max(processes, 1)  # prevent zero processes with 1 CPU

    # Simulate each frame, storing Wigner function data in w_fock
    xvec = numpy.linspace(axes_min, axes_max, axes_steps)

    if keep_state:
        w_fock = __simulate_wigner_with_state(
            circuits, qubit, cbit, xvec, shots, noise_passes, trace
        )
    elif processes == 1:
        w_fock = []
        for circuit in circuits:
            fock, _ = simulate_wigner(
                circuit,
                xvec,
                shots,
                noise_passes=noise_passes,
                conditional=cbit is not None,
                trace=trace or cbit is not None,
            )
            w_fock.append(fock)
    else:
        pool = multiprocessing.Pool(processes)
        results = pool.starmap(
            simulate_wigner,
            (
                (
                    circuit,
                    xvec,
                    shots,
                    noise_passes,
                    cbit is not None,
                    trace or cbit is not None,
                )
                for circuit in circuits
            ),
        )
        pool.close()
        w_fock = [i[0] for i in results if i is not None]

    return w_fock, xvec


def __discretize_wigner_without_measure(
    circuit: CVCircuit,
    animation_segments: int = 10,
    discretize_epsilon: float = None,
    shots: int = 1,
    axes_min: int = -6,
    axes_max: int = 6,
    axes_steps: int = 200,
    noise_passes=None,
    sequential_subcircuit: bool = False,
    trace: bool = True,
):
    statevector_label = "segment_"

    discretized, num_statevectors = discretize_single_circuit(
        circuit=circuit,
        segments_per_gate=animation_segments,
        epsilon=discretize_epsilon,
        sequential_subcircuit=sequential_subcircuit,
        statevector_per_segment=True,
        statevector_label=statevector_label,
        noise_passes=noise_passes,
    )

    xvec = numpy.linspace(axes_min, axes_max, axes_steps)

    w_fock = simulate_wigner_multiple_statevectors(
        circuit=discretized,
        xvec=xvec,
        shots=shots,
        statevector_label=statevector_label,
        num_statevectors=num_statevectors,
        noise_passes=noise_passes,
        trace=trace,
    )

    return w_fock, xvec


def save_animation(anim: matplotlib.animation.FuncAnimation, file: str, bitrate: int):
    file_path = pathlib.Path(file)

    if file_path.suffix == ".mp4":
        writer = matplotlib.animation.FFMpegWriter(fps=24, bitrate=bitrate)
    elif file_path.suffix == ".gif" or file_path.suffix == ".apng":
        writer = matplotlib.animation.PillowWriter(fps=24, bitrate=bitrate)
    else:
        print(
            f"Unknown animation file type {file_path.suffix}, defaulting to using PillowWriter"
        )
        writer = matplotlib.animation.PillowWriter(fps=24, bitrate=bitrate)

    anim.save(file, writer=writer)


def _animate_init():
    pass  # Prevent rendering frame 0 twice (once for init, once for animate)


def _animate(frame, *fargs):
    """Generate individual matplotlib frame in animation."""
    fig = fargs[0]
    ax = fargs[1]
    xvec = fargs[2]
    w_fock = fargs[3][frame]
    file = fargs[4]
    draw_grid = fargs[5]

    amax = numpy.amax(w_fock)
    amin = numpy.amin(w_fock)
    abs_max = max(amax, abs(amin))
    if abs_max == 0:
        abs_max = 5
    color_levels = numpy.linspace(-abs_max, abs_max, 100)

    xvec_int = [int(x) for x in xvec]
    xvec_int = sorted(set(xvec_int))

    ax.clear()
    cont = ax.contourf(xvec, xvec, w_fock, color_levels, cmap="RdBu")

    ax.set_xlabel(r"$x$")
    ax.set_xticks(xvec_int)
    ax.set_ylabel(r"$p$")
    ax.set_yticks(xvec_int)
    if draw_grid:
        ax.grid()

    if frame == 0:
        fig.colorbar(cont, ax=ax)

    time_text = ax.text(
        0.05,
        0.95,
        "",
        horizontalalignment="left",
        verticalalignment="top",
        transform=ax.transAxes,
    )
    time_text.set_text(f"Frame {frame}")

    if file:
        os.makedirs(f"{file}_frames", exist_ok=True)
        plt.savefig(f"{file}_frames/frame_{frame}.png")


def __simulate_wigner_with_state(
    circuits, qubit, cbit, xvec, shots, noise_passes, trace
):
    """Simulate Wigner function, preserving state between iterations"""
    w_fock = []
    previous_state = None
    for circuit in circuits:
        if previous_state:
            # Initialize circuit to simulate with the previous frame's state, then append the last instruction
            sim_circuit = circuit.copy()
            sim_circuit.data.clear()  # Is this safe -- could we copy without data?
            sim_circuit.initialize(previous_state)

            if qubit and cbit:
                last_instructions = circuit.data[
                    -3:
                ]  # Get the last instruction, plus the Hadamard/measure
            else:
                last_instructions = circuit.data[-1:]  # Get the last instruction

            for inst in last_instructions:
                sim_circuit.append(*inst)
        else:
            # No previous simulation state, just run the current circuit
            sim_circuit = circuit
        fock, previous_state = simulate_wigner(
            sim_circuit,
            xvec,
            shots,
            noise_passes=noise_passes,
            conditional=cbit is not None,
            trace=trace or cbit is not None,
        )
        w_fock.append(fock)

    return w_fock



import random
import c2qa
import numpy
import qiskit


def count_nonzero(statevector: qiskit.quantum_info.Statevector):
    """Re-implement numpy.count_nonzero using numpy.isclose()."""
    nonzero = len(statevector.data)
    for state in statevector.data:
        if numpy.isclose(state, 0):
            nonzero -= 1

    return nonzero


def create_conditional(num_qumodes: int = 2, num_qubits_per_qumode: int = 2):
    qmr = c2qa.QumodeRegister(num_qumodes, num_qubits_per_qumode)
    qr = qiskit.QuantumRegister(2)
    circuit = c2qa.CVCircuit(qmr, qr)

    for qumode in range(num_qumodes):
        circuit.cv_initialize(0, qmr[qumode])

    circuit.initialize([0, 1], qr[1])  # qr[0] will init to zero

    return circuit, qmr, qr


def create_unconditional(num_qumodes: int = 2, num_qubits_per_qumode: int = 3):
    qmr = c2qa.QumodeRegister(num_qumodes, num_qubits_per_qumode)
    circuit = c2qa.CVCircuit(qmr)
    for qumode in range(num_qumodes):
        circuit.cv_initialize(0, qmr[qumode])

    return circuit, qmr


def assert_changed(state, result):
    assert result.success
    # print()
    # print(circuit.draw("text"))
    # print(state)

    # TODO - better understand what the state vector results should be
    assert count_nonzero(state) > 1


def assert_unchanged(state, result):
    assert result.success
    # print()
    # print(circuit.draw("text"))
    # print(state)

    # TODO - better understand what the state vector results should be
    assert count_nonzero(state) == 1


def test_no_gates():
    circuit, qmr = create_unconditional()
    state, result, fock_counts = c2qa.util.simulate(circuit)
    assert_unchanged(state, result)


def test_beamsplitter_once():
    circuit, qmr = create_unconditional()

    phi = random.random()
    circuit.cv_bs(phi, qmr[0], qmr[1])

    state, result, fock_counts = c2qa.util.simulate(circuit)

    # TODO - Beam splitter gate does not change state vector
    #        Both Strawberry Fields & FockWits are the same, too.
    # assert_changed(state, result)
    assert_unchanged(state, result)


def test_conditional_beamsplitter():
    circuit, qmr, qr = create_conditional()

    theta = random.random()
    circuit.cv_c_bs(theta, qmr[0], qmr[1], qr[0])

    state, result, fock_counts = c2qa.util.simulate(circuit)

    # TODO - Beam splitter gate does not change state vector
    #        Both Strawberry Fields & FockWits are the same, too.
    # assert_changed(state, result)
    assert_unchanged(state, result)


def test_conditional_schwinger():
    circuit, qmr, qr = create_conditional()

    beta = random.random()
    theta_1 = random.random()
    phi_1 = random.random()
    theta_2 = random.random()
    phi_2 = random.random()
    circuit.cv_c_schwinger(
        [beta, theta_1, phi_1, theta_2, phi_2], qmr[0], qmr[1], qr[0]
    )

    state, result, fock_counts = c2qa.util.simulate(circuit)

    assert_unchanged(state, result)


def test_beamsplitter_twice():
    circuit, qmr = create_unconditional()

    phi = random.random()
    circuit.cv_bs(phi, qmr[0], qmr[1])
    circuit.cv_bs(-phi, qmr[0], qmr[1])

    state, result, fock_counts = c2qa.util.simulate(circuit)
    assert_unchanged(state, result)


def test_conditonal_displacement():
    circuit, qmr, qr = create_conditional()

    alpha = random.random()
    circuit.cv_c_d(alpha, qmr[0], qr[0])
    circuit.cv_c_d(-alpha, qmr[0], qr[0])

    circuit.cv_c_d(-alpha, qmr[0], qr[1])
    circuit.cv_c_d(alpha, qmr[0], qr[1])

    state, result, fock_counts = c2qa.util.simulate(circuit)
    assert_unchanged(state, result)


def test_conditonal_squeezing():
    circuit, qmr, qr = create_conditional()

    alpha = random.random()
    circuit.cv_c_sq(alpha, qmr[0], qr[0])
    circuit.cv_c_sq(-alpha, qmr[0], qr[0])

    circuit.cv_c_sq(-alpha, qmr[0], qr[1])
    circuit.cv_c_sq(alpha, qmr[0], qr[1])

    state, result, fock_counts = c2qa.util.simulate(circuit)
    assert_unchanged(state, result)


def test_displacement_once(capsys):
    with capsys.disabled():
        circuit, qmr = create_unconditional()

        # alpha = random.random()
        alpha = 1
        circuit.cv_d(alpha, qmr[0])

        state, result, fock_counts = c2qa.util.simulate(circuit)
        assert_changed(state, result)


def test_cv_delay():
    circuit, qmr = create_unconditional()

    circuit.cv_delay(100, qmr[0])

    state, result, fock_counts = c2qa.util.simulate(circuit)


def test_displacement_twice():
    circuit, qmr = create_unconditional()

    alpha = random.random()
    circuit.cv_d(alpha, qmr[0])
    circuit.cv_d(-alpha, qmr[0])

    state, result, fock_counts = c2qa.util.simulate(circuit)
    assert_unchanged(state, result)


def test_displacement_calibration(capsys):
    with capsys.disabled():
        qmr = c2qa.QumodeRegister(1, 2)
        qr = qiskit.QuantumRegister(1)
        cr = qiskit.ClassicalRegister(1)
        circuit = c2qa.CVCircuit(qmr, qr, cr)

        # qr[0] and cr[0] will init to zero
        circuit.cv_initialize(0, qmr[0])

        alpha = numpy.sqrt(numpy.pi)

        circuit.h(qr[0])
        circuit.cv_c_d(alpha, qmr[0], qr[0])
        circuit.cv_d(1j * alpha, qmr[0])
        circuit.cv_c_d(-alpha, qmr[0], qr[0])
        circuit.cv_d(-1j * alpha, qmr[0])
        circuit.h(qr[0])
        circuit.measure(qr[0], cr[0])

        state, result, fock_counts = c2qa.util.simulate(circuit)
        assert result.success

        state = result.get_statevector(circuit)
        counts = result.get_counts(circuit)

        assert state.dim > 0
        assert counts

        # print()
        # print(circuit.draw("text"))
        # print(state)
        # print(counts.int_outcomes())


def test_rotation_once():
    circuit, qmr = create_unconditional()

    theta = random.random()
    circuit.cv_r(theta, qmr[0])

    state, result, fock_counts = c2qa.util.simulate(circuit)

    # TODO - Rotation gate does not change state vector.
    #        Both Strawberry Fields & FockWits are the same, too.
    # assert_changed(state, result)
    assert_unchanged(state, result)


def test_rotation_twice():
    circuit, qmr = create_unconditional()

    theta = random.random()
    circuit.cv_r(theta, qmr[0])
    circuit.cv_r(-theta, qmr[0])

    state, result, fock_counts = c2qa.util.simulate(circuit)
    assert_unchanged(state, result)


def test_squeezing_once():
    circuit, qmr = create_unconditional()

    z = random.random()
    circuit.cv_sq(z, qmr[0])

    state, result, fock_counts = c2qa.util.simulate(circuit)
    assert_changed(state, result)


def test_squeezing_twice():
    circuit, qmr = create_unconditional()

    z = random.random()
    circuit.cv_sq(z, qmr[0])
    circuit.cv_sq(-z, qmr[0])

    state, result, fock_counts = c2qa.util.simulate(circuit)
    assert_unchanged(state, result)


def test_two_mode_squeezing_once():
    circuit, qmr = create_unconditional()

    z = random.random()
    circuit.cv_sq2(z, qmr[0], qmr[1])

    state, result, fock_counts = c2qa.util.simulate(circuit)
    assert_changed(state, result)


def test_two_mode_squeezing_twice():
    circuit, qmr = create_unconditional()

    z = random.random()
    circuit.cv_sq2(z, qmr[0], qmr[1])
    circuit.cv_sq2(-z, qmr[0], qmr[1])

    state, result, fock_counts = c2qa.util.simulate(circuit)
    assert_unchanged(state, result)


def test_gates():
    """Verify that we can use the gates, not that they are actually working."""

    # ===== Constants =====
    alpha = 1
    phi = numpy.pi / 2
    z = 1

    circuit, qmr, qr = create_conditional()

    # Basic Gaussian Operations on a Resonator
    circuit.cv_bs(phi, qmr[0], qmr[1])
    circuit.cv_d(alpha, qmr[0])
    circuit.cv_r(phi, qmr[0])
    circuit.cv_sq(z, qmr[0])
    circuit.cv_sq2(z, qmr[0], qmr[1])

    # Hybrid qubit-cavity gates
    circuit.cv_c_d(alpha, qmr[0], qr[0])
    circuit.cv_c_d(alpha, qmr[1], qr[0])
    circuit.cv_c_sq(z, qmr[0], qr[0])
    circuit.cv_c_sq(z, qmr[1], qr[0])

    state, result, fock_counts = c2qa.util.simulate(circuit)

    assert result.success


def test_snap():
    circuit, qmr = create_unconditional()

    phi = random.random()
    n = 1
    circuit.cv_snap(phi, n, qmr[0])

    state, result, fock_counts = c2qa.util.simulate(circuit)


def test_eswap():
    circuit, qmr = create_unconditional()

    phi = random.random()
    circuit.cv_eswap(phi, qmr[0], qmr[1])

    state, result, fock_counts = c2qa.util.simulate(circuit)


def test_multiboson_sampling(capsys):
    with capsys.disabled():
        num_qubits = 1
        num_qumodes = 2
        num_qubits_per_qumode = 2
        qmrA = c2qa.QumodeRegister(
            num_qumodes=num_qumodes,
            num_qubits_per_qumode=num_qubits_per_qumode,
            name="qmrA_initial",
        )
        qmrB = c2qa.QumodeRegister(
            num_qumodes=num_qumodes,
            num_qubits_per_qumode=num_qubits_per_qumode,
            name="qmrB_initial",
        )
        qbr = qiskit.QuantumRegister(size=num_qubits, name="qbr_initial")
        circuit = c2qa.CVCircuit(qmrA, qmrB, qbr)
        circuit.cv_c_multiboson_sampling([0, 1, 2, 3], qmrA[0], qbr[0])

        state, result, fock_counts = c2qa.util.simulate(circuit)
        assert result.success



import numpy
import scipy.sparse
import scipy.sparse.linalg
import pdb


xQB = numpy.array([[0, 1], [1, 0]])
yQB = numpy.array([[0, -1j], [1j, 0]])
zQB = numpy.array([[1, 0], [0, -1]])
idQB = numpy.array([[1, 0], [0, 1]])
sigma_plus = numpy.array([[0, 1], [0, 0]])
sigma_minus = numpy.array([[0, 0], [1, 0]])


class CVOperators:
    """Build operator matrices for continuously variable bosonic gates."""

    def get_a(self, cutoff: int):
        """Annihilation operator"""
        data = numpy.sqrt(range(cutoff))
        return scipy.sparse.spdiags(
            data=data, diags=[1], m=len(data), n=len(data)
        ).tocsc()

    def get_a1(self, cutoff_a: int, cutoff_b: int):
        return scipy.sparse.kron(self.get_a(cutoff_a), self.get_eye(cutoff_b)).tocsc()

    def get_a2(self, cutoff_a: int, cutoff_b: int):
        return scipy.sparse.kron(self.get_eye(cutoff_a), self.get_a(cutoff_b)).tocsc()

    def get_a12(self, cutoff_a: int, cutoff_b: int):
        return self.get_a1(cutoff_a, cutoff_b) * self.get_a2(cutoff_a, cutoff_b)

    def get_a_dag(self, cutoff: int):
        """Creation operator"""
        a = self.get_a(cutoff)
        return a.conjugate().transpose().tocsc()

    def get_N(self, cutoff: int):
        """Number operator"""
        a = self.get_a(cutoff)
        a_dag = self.get_a_dag(cutoff)
        return a_dag * a

    def get_a1_dag(self, cutoff_a: int, cutoff_b: int):
        return self.get_a1(cutoff_a, cutoff_b).conjugate().transpose().tocsc()

    def get_a2_dag(self, cutoff_a: int, cutoff_b: int):
        return self.get_a2(cutoff_a, cutoff_b).conjugate().transpose().tocsc()

    def get_a12_dag(self, cutoff_a: int, cutoff_b: int):
        return self.get_a1_dag(cutoff_a, cutoff_b) * self.get_a2_dag(cutoff_a, cutoff_b)

    def get_a12dag(self, cutoff_a: int, cutoff_b: int):
        return self.get_a1(cutoff_a, cutoff_b) * self.get_a2_dag(cutoff_a, cutoff_b)

    def get_a1dag2(self, cutoff_a: int, cutoff_b: int):
        return self.get_a1_dag(cutoff_a, cutoff_b) * self.get_a2(cutoff_a, cutoff_b)

    def get_eye(self, cutoff: int):
        """Identity matrix"""
        return scipy.sparse.eye(cutoff)

    def r(self, theta, cutoff):
        """Phase space rotation operator

        Args:
            theta (real): rotation

        Returns:
            csc_matrix: operator matrix
        """
        arg = 1j * theta * self.get_N(cutoff)

        return scipy.sparse.linalg.expm(arg)

    def d(self, alpha, cutoff):
        """Displacement operator

        Args:
            alpha (real): displacement

        Returns:
            csc_matrix: operator matrix
        """
        arg = (alpha * self.get_a_dag(cutoff)) - (
            numpy.conjugate(alpha) * self.get_a(cutoff)
        )

        return scipy.sparse.linalg.expm(arg)

    def s(self, theta, cutoff):
        """Single-mode squeezing operator

        Args:
            theta (real): squeeze

        Returns:
            csc_matrix: operator matrix
        """
        a = self.get_a(cutoff)
        a_dag = self.get_a_dag(cutoff)
        a_sqr = a * a
        a_dag_sqr = a_dag * a_dag
        arg = 0.5 * ((numpy.conjugate(theta) * a_sqr) - (theta * a_dag_sqr))

        return scipy.sparse.linalg.expm(arg)

    def s2(self, theta, cutoff_a, cutoff_b):
        """Two-mode squeezing operator

        Args:
            g (real): multiplied by 1j to yield imaginary phase

        Returns:
            csc_matrix: operator matrix
        """

        arg = (numpy.conjugate(theta * 1j) * self.get_a12_dag(cutoff_a, cutoff_b)) - (
            theta * 1j * self.get_a12(cutoff_a, cutoff_b)
        )

        return scipy.sparse.linalg.expm(arg)

    def bs(self, theta, cutoff_a, cutoff_b):
        """Two-mode beam splitter operator

        Args:
            theta: phase

        Returns:
            csc_matrix: operator matrix
        """

        arg = theta * self.get_a1dag2(cutoff_a, cutoff_b) - numpy.conj(
            theta
        ) * self.get_a12dag(cutoff_a, cutoff_b)

        return scipy.sparse.linalg.expm(arg)

    def cr(self, theta, cutoff):
        """Controlled phase space rotation operator

        Args:
            theta (real): phase

        Returns:
            csc_matrix: operator matrix
        """
        arg = theta * 1j * scipy.sparse.kron(zQB, self.get_N(cutoff)).tocsc()

        return scipy.sparse.linalg.expm(arg)

    def crx(self, theta, cutoff):
        """Controlled phase space rotation operator around sigma^x

        Args:
            theta (real): phase

        Returns:
            csc_matrix: operator matrix
        """
        arg = theta * 1j * scipy.sparse.kron(xQB, self.get_N(cutoff)).tocsc()

        return scipy.sparse.linalg.expm(arg)

    def cry(self, theta, cutoff):
        """Controlled phase space rotation operator around sigma^x

        Args:
            theta (real): phase

        Returns:
            csc_matrix: operator matrix
        """
        arg = theta * 1j * scipy.sparse.kron(yQB, self.get_N(cutoff)).tocsc()

        return scipy.sparse.linalg.expm(arg)

    def cd(self, theta, beta, cutoff):
        """Controlled displacement operator

        Args:
            theta (real): displacement for qubit state 0
            beta (real): displacement for qubit state 1. If None, use -alpha.

        Returns:
            bsr_matrix: operator matrix
        """
        displace0 = (theta * self.get_a_dag(cutoff)) - (
            numpy.conjugate(theta) * self.get_a(cutoff)
        )
        if beta is None:
            beta = -theta
        displace1 = (beta * self.get_a_dag(cutoff)) - (
            numpy.conjugate(beta) * self.get_a(cutoff)
        )

        return scipy.sparse.kron(
            (idQB + zQB) / 2, scipy.sparse.linalg.expm(displace0)
        ) + scipy.sparse.kron((idQB - zQB) / 2, scipy.sparse.linalg.expm(displace1))

    def ecd(self, theta, cutoff):
        """Echoed controlled displacement operator

        Args:
            theta (real): displacement

        Returns:
            csr_matrix: operator matrix
        """
        argm = (theta * self.get_a_dag(cutoff)) - (
            numpy.conjugate(theta) * self.get_a(cutoff)
        )
        arg = scipy.sparse.kron(zQB, argm)

        return scipy.sparse.linalg.expm(arg)

    def cbs(self, theta, cutoff_a, cutoff_b):
        """Controlled phase two-mode beam splitter operator

        Args:
            theta (real): real phase

        Returns:
            csc_matrix: operator matrix
        """

        argm = theta * self.get_a1dag2(cutoff_a, cutoff_b) - numpy.conjugate(
            theta
        ) * self.get_a12dag(cutoff_a, cutoff_b)
        arg = scipy.sparse.kron(zQB, argm).tocsc()

        return scipy.sparse.linalg.expm(arg)

    def cschwinger(self, beta, theta_1, phi_1, theta_2, phi_2, cutoff_a, cutoff_b):
        """General form of a controlled Schwinger gate

        Args:
            params (real): [beta, theta_1, phi_1, theta_2, phi_2]

        Returns:
            csc_matrix: operator matrix
        """

        Sx = (
            self.get_a1(cutoff_a, cutoff_b) * self.get_a2_dag(cutoff_a, cutoff_b)
            + self.get_a1_dag(cutoff_a, cutoff_b) * self.get_a2(cutoff_a, cutoff_b)
        ) / 2
        Sy = (
            self.get_a1(cutoff_a, cutoff_b) * self.get_a2_dag(cutoff_a, cutoff_b)
            - self.get_a1_dag(cutoff_a, cutoff_b) * self.get_a2(cutoff_a, cutoff_b)
        ) / (2 * 1j)
        Sz = (
            self.get_a2_dag(cutoff_a, cutoff_b) * self.get_a2(cutoff_a, cutoff_b)
            - self.get_a1_dag(cutoff_a, cutoff_b) * self.get_a1(cutoff_a, cutoff_b)
        ) / 2

        sigma = (
            numpy.sin(theta_1) * numpy.cos(phi_1) * xQB
            + numpy.sin(theta_1) * numpy.sin(phi_1) * yQB
            + numpy.cos(theta_1) * zQB
        )
        S = (
            numpy.sin(theta_2) * numpy.cos(phi_2) * Sx
            + numpy.sin(theta_2) * numpy.sin(phi_2) * Sy
            + numpy.cos(theta_2) * Sz
        )
        arg = scipy.sparse.kron(sigma, S).tocsc()

        return scipy.sparse.linalg.expm(-1j * beta * arg)

    def snap(self, theta, n, cutoff):
        """SNAP (Selective Number-dependent Arbitrary Phase) operator

        Args:
            theta (real): phase
            n (integer): Fock state in which the mode should acquire the phase

        Returns:
            csc_matrix: operator matrix
        """

        ket_n = numpy.zeros(cutoff)
        ket_n[n] = 1
        projector = numpy.outer(ket_n, ket_n)
        sparse_projector = scipy.sparse.csr_matrix(projector)
        arg = theta * 1j * sparse_projector.tocsc()
        return scipy.sparse.linalg.expm(arg)

    def csnap(self, theta, n, cutoff):
        """SNAP (Selective Number-dependent Arbitrary Phase) operator,
        with explicit sigma_z in exponential. Can be used to generate
        fock-number selective qubit rotations.

        Args:
            theta (real): phase
            n (integer): Fock state in which the mode should acquire the phase

        Returns:
            csc_matrix: operator matrix
        """

        ket_n = numpy.zeros(cutoff)
        ket_n[n] = 1
        projector = numpy.outer(ket_n, ket_n)
        # sparse_projector = scipy.sparse.csc_matrix(projector)
        arg = theta * 1j * scipy.sparse.kron(zQB, projector).tocsc()
        return scipy.sparse.linalg.expm(arg)

    def multisnap(self, *args):
        """SNAP (Selective Number-dependent Arbitrary Phase) operator for multiple Fock states.
        Generates an arbitrary number of fock-number selective qubit rotations.

        Args:
            args (List[reals, integers]): [List of phases, List of Fock states in which the mode should acquire the associated phase]

        Returns:
            csr_matrix: operator matrix
        """
        # Divide list in two because thetas and ns must be sent in as a single list
        cutoff = args[-1]
        theta_ns = args[0:-1]
        thetas = theta_ns[: len(theta_ns) // 2]  # arguments
        ns = theta_ns[len(theta_ns) // 2 :]  # Fock states on which they are applied
        if len(thetas) != len(ns):  # one theta per Fock state to apply it to
            raise Exception("len(theta) must be equal to len(n)")

        id = numpy.eye(cutoff)
        gate = scipy.sparse.csr_matrix(id)
        for i in range(len(ns)):
            ket_n = numpy.zeros(cutoff)
            ket_n[ns[i]] = 1
            projector = numpy.outer(ket_n, ket_n)
            coeff = numpy.exp(1j * thetas[i]) - 1
            mat = scipy.sparse.csr_matrix(coeff * projector)
            gate = numpy.add(gate, mat)
        return scipy.sparse.csr_matrix(gate)

    def multicsnap(self, *args):
        """SNAP (Selective Number-dependent Arbitrary Phase) operator for multiple Fock states.
        Generates an arbitrary number of fock-number selective qubit rotations, with the qubit that accrues the geometric phase explicit.

        Args:
            args (List[reals, integers]): [List of phases, List of Fock states in which the mode should acquire the associated phase]

        Returns:
            csr_matrix: operator matrix
        """
        # Divide list in two because thetas and ns must be sent in as a single list
        cutoff = args[-1]
        theta_ns = args[0:-1]
        thetas = theta_ns[: len(theta_ns) // 2]  # arguments
        ns = theta_ns[len(theta_ns) // 2 :]  # Fock states on which they are applied
        if len(thetas) != len(ns):  # one theta per Fock state to apply it to
            raise Exception("len(theta) must be equal to len(n)")

        id = numpy.eye(cutoff)
        gate = scipy.sparse.csr_matrix(scipy.sparse.kron(idQB, id))
        for i in range(len(ns)):
            ket_n = numpy.zeros(cutoff)
            ket_n[ns[i]] = 1
            projector = numpy.outer(ket_n, ket_n)
            coeff = scipy.sparse.linalg.expm(1j * thetas[i] * zQB) - idQB
            mat = scipy.sparse.kron(coeff, projector).tocsr()
            gate = numpy.add(gate, mat)
        return scipy.sparse.csr_matrix(gate)

    def pnr(self, max, cutoff):
        """Support gate for photon number readout (see Curtis et al., PRA (2021) and Wang et al., PRX (2020))

        Args:
            max (int): the period of the mapping

        Returns:
            csc_matrix: operator matrix
        """
        ket_n = numpy.zeros(cutoff)
        projector = numpy.outer(ket_n, ket_n)
        # binary search
        for j in range(int(max / 2)):
            for i in range(j, cutoff, max):
                ket_n = numpy.zeros(cutoff)
                # fill from right to left
                ket_n[-(i + 1)] = 1
                projector += numpy.outer(ket_n, ket_n)

        # Flip qubit if there is a boson present in any of the modes addressed by the projector
        arg = 1j * (-numpy.pi / 2) * scipy.sparse.kron(xQB, projector).tocsc()
        return scipy.sparse.linalg.expm(arg)

    def eswap(self, theta, cutoff_a, cutoff_b):
        """Exponential SWAP operator

        Args:
            theta (real): rotation

        Returns:
            csc_matrix: operator matrix
        """

        self.mat = numpy.zeros([cutoff_a * cutoff_b, cutoff_a * cutoff_b])
        for j in range(cutoff_a):
            for i in range(cutoff_b):
                self.mat[i + (j * cutoff_a)][i * cutoff_b + j] = 1
        self.sparse_mat = scipy.sparse.csr_matrix(self.mat).tocsc()

        arg = 1j * theta * self.sparse_mat

        return scipy.sparse.linalg.expm(arg)

    def csq(self, theta, cutoff):
        """Single-mode squeezing operator

        Args:
            theta (real): squeeze

        Returns:
            csc_matrix: operator matrix
        """
        a_sqr = self.get_a(cutoff) * self.get_a(cutoff)
        a_dag_sqr = self.get_a_dag(cutoff) * self.get_a_dag(cutoff)
        arg = scipy.sparse.kron(
            zQB, 0.5 * ((numpy.conjugate(theta) * a_sqr) - (theta * a_dag_sqr))
        ).tocsc()

        return scipy.sparse.linalg.expm(arg)

    def testqubitorderf(self, phi):

        arg = 1j * phi * scipy.sparse.kron(xQB, idQB)
        return scipy.sparse.linalg.expm(arg)

    def c_multiboson_sampling(self, max, cutoff):
        """SNAP gate creation for multiboson sampling purposes.

        Args:
            max (int): the period of the mapping

        Returns:
            dia_matrix: operator matrix
        """
        print(max)

        return self.get_eye(cutoff)

    def gate_from_matrix(self, matrix):
        """Converts matrix into gate. Called using ParameterizedUnitaryGate.

        Args:
            matrix (list): the (unitary) matrix that you wish to convert into a gate

        Returns:
            csc_matrix: operator matrix
        """
        return scipy.sparse.csc_matrix(matrix)



from pathlib import Path


import numpy
import pytest
import qiskit


import c2qa


def __build_subcircuit():
    # Define Hamiltonian parameters
    omega_R = 2
    omega_Q = 5
    chi = 0.1

    # Set number of qubits per qumode
    num_qubits_per_qumode = 3

    # Choose alpha for coherent state
    alpha = 1

    # Choose total animation time
    total_time = 1 * 2 * numpy.pi / omega_R

    # Create new circuit
    qmr = c2qa.QumodeRegister(
        num_qumodes=1, num_qubits_per_qumode=num_qubits_per_qumode
    )
    qbr = qiskit.QuantumRegister(1)
    U_JC = c2qa.CVCircuit(qmr, qbr)

    # Append U_R
    U_JC.cv_r(-omega_R * total_time, qmr[0])
    # Append U_Q
    U_JC.rz(omega_Q * total_time, qbr[0])
    # Append U_\chi -- KS: this needs to be updated to reflect naming conventions in manuscript
    U_JC.cv_c_r(-chi * total_time / 2, qmr[0], qbr[0])
    # Compile this circuit into a single parameterized gate
    U_JC = U_JC.to_gate(label="U_JC")

    # Instantiate the circuit and initialize the qubit to the '0' state.
    circuit_0 = c2qa.CVCircuit(qmr, qbr)
    circuit_0.initialize([1, 0], qbr)

    # Squeeze so we can visually see rotation
    circuit_0.cv_sq(0.5, qmr[0])

    # Now initialize the qumode in a coherent state
    # cutoff = 2**num_qubits_per_qumode
    # coeffs = [numpy.exp(-numpy.abs(alpha)**2/2)*alpha**n/(numpy.sqrt(numpy.math.factorial(n))) for n in range(0,cutoff)]
    # circuit_0.cv_initialize(coeffs,qmr[0])

    # Append time evolution unitary
    circuit_0.append(U_JC, qmr[0] + [qbr[0]])
    # circuit_0.assign_parameters({dt : total_time})

    # dt = total_time
    # # Append U_R
    # circuit_0.cv_r(-omega_R*dt,qmr[0])
    # # Append U_Q
    # circuit_0.rz(omega_Q*dt,qbr[0])
    # # Append U_\chi -- KS: this needs to be updated to reflect naming conventions in manuscript
    # circuit_0.cv_c_r(-chi*dt/2,qmr[0],qbr[0])

    # Compile this circuit into a single parameterized gate
    # U_JC = U_JC.to_gate(label='U_JC')

    # # Now repeat the above steps for a qubit initialized to the '1' state:
    # circuit_1 = c2qa.CVCircuit(qmr,qbr)
    # circuit_1.initialize([0,1], qbr)
    # circuit_1.cv_d(alpha,qmr[0])
    # circuit_1.append(U_JC,qmr[0] + [qbr[0]])
    # circuit_1 = circuit_1.assign_parameters({dt : total_time})

    return circuit_0


def test_animate_subcircuit_one_gate(capsys):
    """Test animating a circuit with a composite gate built from another circuit.
    Composite gate borrowed from Jaynes-Cummings model tutorial"""

    with capsys.disabled():
        circuit = __build_subcircuit()

        # Animate wigner function of each circuit
        c2qa.animate.animate_wigner(
            circuit, file="tests/composite_gate.gif", animation_segments=20
        )


def test_animate_subcircuit_sequential(capsys):
    """Test animating a circuit with a composite gate built from another circuit.
    Composite gate borrowed from Jaynes-Cummings model tutorial"""

    with capsys.disabled():
        circuit = __build_subcircuit()

        # Animate wigner function of each circuit
        c2qa.animate.animate_wigner(
            circuit,
            file="tests/sequential_subcircuit.gif",
            animation_segments=20,
            sequential_subcircuit=True,
        )


def test_animate_parameterized(capsys):
    with capsys.disabled():
        a = qiskit.circuit.Parameter("𝛼")

        qmr = c2qa.QumodeRegister(1, num_qubits_per_qumode=4)
        qbr = qiskit.QuantumRegister(1)
        cbr = qiskit.ClassicalRegister(1)

        minimal_circuit = c2qa.CVCircuit(qmr, qbr, cbr)

        minimal_circuit.h(qbr[0])

        minimal_circuit.cv_c_d(1j * a, qmr[0], qbr[0])

        bound_circuit = minimal_circuit.assign_parameters({a: 2})

        wigner_filename = "tests/animate_parameterized.apng"
        c2qa.animate.animate_wigner(
            bound_circuit,
            qubit=qbr[0],
            cbit=cbr[0],
            file=wigner_filename,
            axes_min=-8,
            axes_max=8,
            animation_segments=5,
            processes=1,
            shots=25,
        )
        assert Path(wigner_filename).is_file()


def test_animate_gif(capsys):
    with capsys.disabled():
        __animate_with_cbit("tests/displacement.gif")


def test_animate_apng(capsys):
    with capsys.disabled():
        __animate_with_cbit("tests/displacement.apng")


def __animate_with_cbit(filename: str):
    qmr = c2qa.QumodeRegister(num_qumodes=1, num_qubits_per_qumode=4)
    qr = qiskit.QuantumRegister(size=1)
    cr = qiskit.ClassicalRegister(size=1)
    circuit = c2qa.CVCircuit(qmr, qr, cr)

    dist = 3

    circuit.initialize([1, 0], qr[0])
    circuit.cv_initialize(0, qmr[0])

    circuit.h(qr[0])
    circuit.cv_c_d(dist, qmr[0], qr[0])

    c2qa.animate.animate_wigner(
        circuit,
        qubit=qr[0],
        cbit=cr[0],
        file=filename,
        axes_min=-8,
        axes_max=8,
        animation_segments=5,
        processes=1,
        shots=25,
    )
    assert Path(filename).is_file()


def __animate_without_cbit(filename: str, trace: bool = False):
    qmr = c2qa.QumodeRegister(num_qumodes=1, num_qubits_per_qumode=4)
    qr = qiskit.QuantumRegister(size=1)
    circuit = c2qa.CVCircuit(qmr, qr)

    dist = 3

    circuit.initialize([1, 0], qr[0])
    circuit.cv_initialize(0, qmr[0])

    circuit.h(qr[0])
    circuit.cv_c_d(dist, qmr[0], qr[0])

    c2qa.animate.animate_wigner(
        circuit,
        qubit=qr[0],
        file=filename,
        axes_min=-8,
        axes_max=8,
        animation_segments=5,
        processes=1,
        shots=25,
        trace=trace,
    )
    assert Path(filename).is_file()


def test_animate_with_trace(capsys):
    with capsys.disabled():
        __animate_without_cbit("tests/animate_with_trace.gif", True)


def test_animate_without_trace(capsys):
    with capsys.disabled():
        __animate_without_cbit("tests/animate_without_trace.gif", False)


def test_calibration_animate_mp4(capsys):
    with capsys.disabled():
        qmr = c2qa.QumodeRegister(num_qumodes=1, num_qubits_per_qumode=6)
        qr = qiskit.QuantumRegister(size=1)
        cr = qiskit.ClassicalRegister(size=1)
        circuit = c2qa.CVCircuit(qmr, qr, cr)

        dist = 3

        circuit.initialize([1, 0], qr[0])
        circuit.cv_initialize(0, qmr[0])

        circuit.h(qr[0])
        circuit.cv_c_d(dist, qmr[0], qr[0])
        circuit.cv_d(1j * dist, qmr[0])
        circuit.cv_c_d(-dist, qmr[0], qr[0])
        circuit.cv_d(-1j * dist, qmr[0])

        c2qa.animate.animate_wigner(
            circuit,
            qubit=qr[0],
            cbit=cr[0],
            file="tests/displacement.gif",
            axes_min=-8,
            axes_max=8,
            animation_segments=48,
            shots=128,
        )



import math
import qiskit

from c2qa.circuit import CVCircuit
from c2qa.kraus import PhotonLossNoisePass
from c2qa.parameterized_unitary_gate import ParameterizedUnitaryGate


def discretize_circuits(
    circuit: CVCircuit,
    segments_per_gate: int = 10,
    keep_state: bool = True,
    qubit: qiskit.circuit.quantumcircuit.QubitSpecifier = None,
    cbit: qiskit.circuit.quantumcircuit.QubitSpecifier = None,
    sequential_subcircuit: bool = False,
):
    """
    Discretize gates into a circuit into segments where each segment ends an indiviudal circuit. Useful for incrementally applying noise or animating the circuit.

    Args:
        circuit (CVCircuit): circuit to simulate and plot
        segments_per_gate (int, optional): Number of segments to split each gate into. Defaults to 10.
        keep_state (bool, optional): True if each gate segments builds on the previous gegment's state vector.
                                     False if each segment starts over from the beginning of the circuit.
                                     If True, it requires sequential simulation of each segment.
        qubit ([QubitSpecifier]): Qubit to measure, if performing Hadamard measure for use with cat states. Defaults to None.
        cbit ([QubitSpecifier]): Classical bit to measure into, if performing Hadamard measure for use with cat states. Defaults to None.
        sequential_subcircuit (bool, optional): boolean flag to animate subcircuits as one gate (False) or as sequential
                                                gates (True). Defautls to False.

    Returns:
        [list]: List of discretized Qiskit circuit
    """

    sim_circuits = []  # Each segment will have its own circuit to simulate

    # base_circuit is copied each gate iteration to build circuit segments to simulate
    base_circuit = circuit.copy()
    base_circuit.data.clear()  # Is this safe -- could we copy without data?

    for inst, qargs, cargs in circuit.data:
        # TODO - get qubit & cbit for measure instead of using parameters
        # qubit = xxx
        # cbit = yyy

        segments = __to_segments(
            inst, segments_per_gate, keep_state, sequential_subcircuit
        )

        for segment in segments:
            sim_circuit = base_circuit.copy()

            sim_circuit.append(instruction=segment, qargs=qargs, cargs=cargs)

            if qubit and cbit:
                # sim_circuit.barrier()
                sim_circuit.h(qubit)
                sim_circuit.measure(qubit, cbit)

            sim_circuits.append(sim_circuit)

            # Start with current circuit for the next segment
            base_circuit = sim_circuit

    return sim_circuits


def discretize_single_circuit(
    circuit: CVCircuit,
    segments_per_gate: int = 10,
    epsilon: float = None,
    sequential_subcircuit: bool = False,
    statevector_per_segment: bool = False,
    statevector_label: str = "segment_",
    noise_passes=None,
):
    """
    Discretize gates into a circuit into segments within a single output circuit. Useful for incrementally applying noise or animating the circuit.

    Args:
        circuit (CVCircuit): circuit to simulate and plot
        segments_per_gate (int, optional): Number of segments to split each gate into. Defaults to 10.
        epsilon (float, optional): float value used to discretize, must specify along with kappa
        kappa (float, optional): float phton loss rate to determine discretization sice, must specify along with epsilon
        sequential_subcircuit (bool, optional): boolean flag to animate subcircuits as one gate (False) or as sequential
                                                gates (True). Defaults to False.
        statevector_per_segment (bool, optional): boolean flag to save a statevector per gate segment. True will call Qiskit
                                                  save_statevector after each segment is simulated, creating statevectors labeled
                                                  "segment_*" that can used after simulation. Defaults to False.
        statevector_label (str, optional): String prefix to use for the statevector saved after each segment
        noise_passes (list of Qiskit noise passes, optional): noise passes to apply

    Returns:
        discretized Qiskit circuit
    """

    # discretized is a copy of the circuit as a whole. Each gate segment be added to simulate
    discretized = circuit.copy()
    discretized.data.clear()  # Is this safe -- could we copy without data?

    if noise_passes:
        if not isinstance(noise_passes, list):
            noise_passes = [noise_passes]

    segment_count = 0
    for inst, qargs, cargs in circuit.data:
        num_segments = segments_per_gate
        qargs_indices = [
            qubit._index for qubit in qargs
        ]  # FIXME -- is there a public API to get the qubit's index in Qiskit v1.0+?

        if noise_passes and not (
            isinstance(inst, qiskit.circuit.instruction.Instruction)
            and inst.name == "initialize"
        ):  # Don't discretize instructions initializing system state:
            noise_pass = None
            for current in noise_passes:
                if isinstance(
                    current, PhotonLossNoisePass
                ) and current.applies_to_instruction(inst, qargs_indices):
                    noise_pass = current
                    break

            if epsilon is not None and noise_pass is not None:
                # FIXME - which of the qumodes' loss rates and QumodeRegister's cutoff should we use?
                photon_loss_rate = noise_pass.photon_loss_rates_sec[0]
                num_segments = math.ceil(
                    (
                        photon_loss_rate
                        * noise_pass.duration_to_sec(inst)
                        * circuit.get_qmr_cutoff(0)
                    )
                    / epsilon
                )

        segments = __to_segments(
            inst=inst,
            segments_per_gate=num_segments,
            keep_state=True,
            sequential_subcircuit=sequential_subcircuit,
        )

        for segment in segments:
            discretized.append(instruction=segment, qargs=qargs, cargs=cargs)

            if statevector_per_segment:
                discretized.save_statevector(
                    label=f"{statevector_label}{segment_count}"
                )
                segment_count += 1

    return discretized, segment_count


def __to_segments(
    inst: qiskit.circuit.instruction.Instruction,
    segments_per_gate: int,
    keep_state: bool,
    sequential_subcircuit: bool,
):
    """Split the instruction into segments_per_gate segments"""

    if isinstance(inst, ParameterizedUnitaryGate):
        # print(f"Discretizing ParameterizedUnitaryGate {inst.name}")
        segments = __discretize_parameterized(inst, segments_per_gate, keep_state)

    # FIXME -- how to identify a gate that was made with QuantumCircuit.to_gate()?
    elif (
        isinstance(inst.definition, qiskit.QuantumCircuit)
        and inst.name != "initialize"
        and inst.label != "cv_gate_from_matrix"
        and len(inst.decompositions) == 0
    ):  # Don't animate subcircuits initializing system state
        # print(f"Discretizing QuantumCircuit {inst.name}")
        segments = __discretize_subcircuit(
            inst.definition, segments_per_gate, keep_state, sequential_subcircuit
        )

    elif (
        isinstance(inst, qiskit.circuit.instruction.Instruction)
        and inst.name != "initialize"
        and inst.label != "cv_gate_from_matrix"
        and len(inst.params) > 0
    ):  # Don't animate instructions initializing system state
        # print(f"Discretizing Instruction {inst.name}")
        segments = __discretize_instruction(inst, segments_per_gate, keep_state)

    else:
        # Else just "discretize" the instruction as a single segment
        # print(f"NOT discretizing {inst.name}")
        segments = [inst]

    return segments


def __discretize_parameterized(
    inst: qiskit.circuit.instruction.Instruction,
    segments_per_gate: int,
    keep_state: bool,
    discretized_param_indices: list = [],
):
    """Split ParameterizedUnitaryGate into multiple segments"""
    segments = []
    for index in range(1, segments_per_gate + 1):
        params = inst.calculate_segment_params(
            current_step=index,
            total_steps=segments_per_gate,
            keep_state=keep_state,
        )
        duration, unit = inst.calculate_segment_duration(
            current_step=index,
            total_steps=segments_per_gate,
            keep_state=keep_state,
        )

        # print(f"Discretized params {params} duration {duration} unit {unit}")

        segments.append(
            ParameterizedUnitaryGate(
                inst.op_func,
                params=params,
                cutoffs=inst.cutoffs,
                num_qubits=inst.num_qubits,
                label=inst.label,
                duration=duration,
                unit=unit,
            )
        )

    return segments


def __discretize_subcircuit(
    subcircuit: qiskit.QuantumCircuit,
    segments_per_gate: int,
    keep_state: bool,
    sequential_subcircuit: bool,
):
    """Create a list of circuits where the entire subcircuit is converted into segments (vs a single instruction)."""

    segments = []
    sub_segments = []

    for inst, qargs, cargs in subcircuit.data:
        sub_segments.append(
            (
                __to_segments(
                    inst, segments_per_gate, keep_state, sequential_subcircuit
                ),
                qargs,
                cargs,
            )
        )

    if sequential_subcircuit:
        # Sequentially animate each gate within the subcircuit
        subcircuit_copy = subcircuit.copy()
        subcircuit_copy.data.clear()  # Is this safe -- could we copy without data?

        for sub_segment in sub_segments:
            gates, gate_qargs, gate_cargs = sub_segment
            for gate in gates:
                subcircuit_copy.append(gate, gate_qargs, gate_cargs)

        segments.append(subcircuit)
    else:
        # Animate the subcircuit as one gate
        for segment in range(segments_per_gate):
            subcircuit_copy = subcircuit.copy()
            subcircuit_copy.data.clear()  # Is this safe -- could we copy without data?

            for sub_segment, qargs, cargs in sub_segments:
                subcircuit_copy.append(sub_segment[segment], qargs, cargs)

            segments.append(subcircuit_copy)

    return segments


def __discretize_instruction(
    inst: qiskit.circuit.instruction.Instruction,
    segments_per_gate: int,
    keep_state: bool,
):
    """Split Qiskit Instruction into multiple segments"""
    segments = []

    for index in range(1, segments_per_gate + 1):
        params = inst.calculate_segment_params(
            current_step=index,
            total_steps=segments_per_gate,
            keep_state=keep_state,
        )
        duration, unit = inst.calculate_segment_duration(
            current_step=index,
            total_steps=segments_per_gate,
            keep_state=keep_state,
        )

        segments.append(
            qiskit.circuit.instruction.Instruction(
                name=inst.name,
                num_qubits=inst.num_qubits,
                num_clbits=inst.num_clbits,
                params=params,
                duration=duration,
                unit=unit,
                label=inst.label,
            )
        )

    return segments



import c2qa
import json
import numpy
import pytest
import qiskit
from qiskit_ibm_runtime.utils import RuntimeEncoder


def test_no_registers():
    with pytest.raises(ValueError):
        c2qa.CVCircuit()


def test_only_quantumregister():
    with pytest.raises(ValueError):
        qr = qiskit.QuantumRegister(1)
        c2qa.CVCircuit(qr)


def test_only_qumoderegister():
    c2qa.CVCircuit(c2qa.QumodeRegister(1, 1))


def test_correct():
    c2qa.CVCircuit(qiskit.QuantumRegister(1), c2qa.QumodeRegister(1, 1))


def test_with_classical():
    c2qa.CVCircuit(
        qiskit.QuantumRegister(1),
        c2qa.QumodeRegister(1, 1),
        qiskit.ClassicalRegister(1),
    )


def test_with_initialize():
    number_of_modes = 5
    number_of_qubits = number_of_modes
    number_of_qubits_per_mode = 2

    qmr = c2qa.QumodeRegister(
        num_qumodes=number_of_modes, num_qubits_per_qumode=number_of_qubits_per_mode
    )
    qbr = qiskit.QuantumRegister(size=number_of_qubits)
    cbr = qiskit.ClassicalRegister(size=1)
    circuit = c2qa.CVCircuit(qmr, qbr, cbr)

    sm = [0, 0, 1, 0, 0]
    for i in range(qmr.num_qumodes):
        circuit.cv_initialize(sm[i], qmr[i])

    circuit.initialize(numpy.array([0, 1]), qbr[0])

    state, result, _ = c2qa.util.simulate(circuit)
    assert result.success


def test_with_delay(capsys):
    with capsys.disabled():
        number_of_modes = 1
        number_of_qubits = 1
        number_of_qubits_per_mode = 2

        qmr = c2qa.QumodeRegister(
            num_qumodes=number_of_modes, num_qubits_per_qumode=number_of_qubits_per_mode
        )
        qbr = qiskit.QuantumRegister(size=number_of_qubits)
        circuit = c2qa.CVCircuit(qmr, qbr)

        circuit.delay(100)
        circuit.cv_d(1, qmr[0])

        state, result, _ = c2qa.util.simulate(circuit)
        assert result.success


def test_get_qubit_indices(capsys):
    with capsys.disabled():
        number_of_modes = 2
        number_of_qubits = 2
        number_of_qubits_per_mode = 2

        qmr = c2qa.QumodeRegister(
            num_qumodes=number_of_modes, num_qubits_per_qumode=number_of_qubits_per_mode
        )
        qbr = qiskit.QuantumRegister(size=number_of_qubits)
        cbr = qiskit.ClassicalRegister(size=1)
        circuit = c2qa.CVCircuit(qmr, qbr, cbr)

        indices = circuit.get_qubit_indices([qmr[1]])
        print(f"qmr[1] indices = {indices}")
        assert indices == [2, 3]

        indices = circuit.get_qubit_indices([qmr[0]])
        print(f"qmr[0] indices = {indices}")
        assert indices == [0, 1]

        indices = circuit.get_qubit_indices([qbr[1]])
        print(f"qbr[1] indices = {indices}")
        assert indices == [5]

        indices = circuit.get_qubit_indices([qbr[0]])
        print(f"qbr[0] indices = {indices}")
        assert indices == [4]


def test_initialize_qubit_values(capsys):
    with capsys.disabled():
        print()

        number_of_modes = 1
        number_of_qubits_per_mode = 4

        for fock in range(pow(2, number_of_qubits_per_mode)):
            qmr = c2qa.QumodeRegister(
                num_qumodes=number_of_modes,
                num_qubits_per_qumode=number_of_qubits_per_mode,
            )
            circuit = c2qa.CVCircuit(qmr)
            circuit.cv_initialize(fock, qmr[0])

            state, result, _ = c2qa.util.simulate(circuit)
            assert result.success

            print(f"fock {fock} qubits {list(result.get_counts().keys())[0]}")


def test_serialize(capsys):
    with capsys.disabled():
        print()

        qumodes = c2qa.QumodeRegister(2)
        bosonic_circuit = c2qa.CVCircuit(qumodes)

        init_state = [0, 2]
        for i in range(qumodes.num_qumodes):
            bosonic_circuit.cv_initialize(init_state[i], qumodes[i])

        phi = qiskit.circuit.Parameter("phi")
        bosonic_circuit.cv_bs(phi, qumodes[0], qumodes[1])

        # print(bosonic_circuit.draw())
        print("\nAttempt to serialize an unbound CVCircuit:")
        bosonic_serial = json.dumps(bosonic_circuit, cls=RuntimeEncoder)
        print(bosonic_serial)


def test_cv_gate_from_matrix(capsys):
    with capsys.disabled():
        # This test picks random qumode/qubit to initialize xgate on, does fockcounts, and checks that results match expected values.
        xgate = [
            [0, 1],
            [1, 0],
        ]  # For qubit, flips |1> to |0> and vice versa. For qumode, flips fock |1> to fock |0> and vice versa

        num_qumode_registers = 2
        num_qumodes_per_register = 2
        num_qubits_per_qumode = 1

        num_qubits = 3

        total_qubits = (
            num_qumode_registers * num_qumodes_per_register * num_qubits_per_qumode
            + num_qubits
        )
        for i in range(10):  # Repeat test 10 times
            # Two qumode registers, One quantum register, One classical register. Hilbert space dimension = 2**(2 * 2 * 1 + 3) = 128
            qmr1 = c2qa.QumodeRegister(num_qumodes_per_register, num_qubits_per_qumode)
            qmr2 = c2qa.QumodeRegister(num_qumodes_per_register, num_qubits_per_qumode)
            q = qiskit.QuantumRegister(num_qubits)
            creg = qiskit.ClassicalRegister(total_qubits)

            # Initialize all qumodes to fock |1> and all qubits to |1> state
            circuit = c2qa.CVCircuit(qmr1, qmr2, q, creg)
            circuit.cv_initialize([0, 1], qmr1)
            circuit.cv_initialize([0, 1], qmr2)

            for i in range(num_qubits):
                circuit.initialize([0, 1], q[i])

            # Pick one of the qumode registers to act on
            register = numpy.random.randint(1, 3)

            # Pick one of the qumodes to act on
            qumode_no = numpy.random.randint(0, 2)

            # Pick one of the qubits to act on
            qubit_no = numpy.random.randint(0, 3)

            # Create string corresponding to expected results
            expect = ["1" for _ in range(total_qubits)]

            if register == 1:
                expect[qumode_no] = "0"
            elif register == 2:
                expect[qumode_no + 2] = "0"

            expect[4 + qubit_no] = "0"

            expect = "".join(reversed(expect))

            # Depending on quantum register chosen, assert result to be true
            if register == 1:
                circuit.cv_gate_from_matrix(xgate, qumodes=qmr1[qumode_no])
                circuit.cv_gate_from_matrix(xgate, qubits=q[qubit_no])

                circuit.cv_measure(qmr1[:] + qmr2[:] + q[:], creg)

                _, result, fock_counts = c2qa.util.simulate(circuit)

                # There should only be 1 result
                if len(list(fock_counts.keys())) > 1:
                    raise Exception

                assert list(fock_counts.keys())[0] == expect

            elif register == 2:
                circuit.cv_gate_from_matrix(xgate, qmr2[qumode_no])
                circuit.cv_gate_from_matrix(xgate, qubits=q[qubit_no])

                circuit.cv_measure(qmr1[:] + qmr2[:] + q[:], creg)

                _, result, fock_counts = c2qa.util.simulate(circuit)

                # There should only be 1 result
                if len(list(fock_counts.keys())) > 1:
                    raise Exception

                assert list(fock_counts.keys())[0] == expect

            else:
                raise Exception


def test_discretize_cv_gate_from_matrix(capsys):
    with capsys.disabled():
        # This test picks random qumode/qubit to initialize xgate on, does fockcounts, and checks that results match expected values.
        xgate = [[0, 1], [1, 0]]

        num_qumode_registers = 2
        num_qumodes_per_register = 2
        num_qubits_per_qumode = 1

        num_qubits = 3

        total_qubits = (
            num_qumode_registers * num_qumodes_per_register * num_qubits_per_qumode
            + num_qubits
        )

        # Two qumode registers, One quantum register, One classical register. Hilbert space dimension = 2**(2 * 2 * 1 + 3) = 128
        qmr1 = c2qa.QumodeRegister(num_qumodes_per_register, num_qubits_per_qumode)
        qmr2 = c2qa.QumodeRegister(num_qumodes_per_register, num_qubits_per_qumode)
        q = qiskit.QuantumRegister(num_qubits)
        creg = qiskit.ClassicalRegister(total_qubits)

        # Initialize all qumodes to fock |1> and all qubits to |1> state
        circuit = c2qa.CVCircuit(qmr1, qmr2, q, creg)

        circuit.cv_gate_from_matrix(xgate, qumodes=qmr1[0])

        discretized = c2qa.discretize.discretize_circuits(circuit)
        assert len(circuit.data) == len(discretized)


def test_cv_initialize(capsys):
    with capsys.disabled():
        qmr1 = c2qa.QumodeRegister(1, 2)
        qmr2 = c2qa.QumodeRegister(2, 2)
        qmr3 = c2qa.QumodeRegister(2, 3)  # <----- change to (2, 2) for no error
        qmr4 = c2qa.QumodeRegister(1, 2)
        qmr5 = c2qa.QumodeRegister(3, 2)
        qmr6 = c2qa.QumodeRegister(3, 2)
        circuit = c2qa.CVCircuit(qmr1, qmr2, qmr3, qmr4, qmr5, qmr6)

        circuit.cv_initialize([0, 1], qmr1[0])
        circuit.cv_initialize([0, 1], qmr2[0])
        circuit.cv_initialize([0, 1], qmr3[0])
        circuit.cv_initialize([0, 1], qmr4[0])
        circuit.cv_initialize([0, 1], qmr5[0])
        circuit.cv_initialize([0, 1], qmr6[0])

        # saving a state vector for all the registers takes a considerable amount of time
        state, result, fock_counts = c2qa.util.simulate(
            circuit, add_save_statevector=False, return_fockcounts=False
        )
        assert result.success



from qiskit import QuantumRegister


class QumodeRegister:
    """Wrapper to QisKit QuantumRegister to represent multiple qubits per qumode.

    Implements __getitem__ to make QumodeRegister appear to work just like QuantumRegister with instances of CVCircuit.
    """

    def __init__(
        self, num_qumodes: int, num_qubits_per_qumode: int = 2, name: str = None
    ):
        """Initialize QumodeRegister

        Args:
            num_qumodes (int): total number of qumodes
            num_qubits_per_qumode (int, optional): Number of qubits representing each qumode. Defaults to 2.
            name (str, optional): Name of register. Defaults to None.
        """
        self.size = num_qumodes * num_qubits_per_qumode
        self.num_qumodes = num_qumodes
        self.num_qubits_per_qumode = num_qubits_per_qumode
        self.cutoff = QumodeRegister.calculate_cutoff(num_qubits_per_qumode)

        # Aggregate the QuantumRegister representing these qumodes as
        # extending the class confuses QisKit when overriding __getitem__().
        # It doesn't expect a list of Qubit back when indexing a single value
        # (i.e., qmr[0] is represented by multiple qubits).
        self.qreg = QuantumRegister(size=self.size, name=name)

    @staticmethod
    def calculate_cutoff(num_qubits_per_qumode: int):
        return 2**num_qubits_per_qumode

    def get_qumode_index(self, qubit):
        """Get the qumode index for the given qubit in this register"""
        qubit_index = self.qreg.index(qubit)
        return qubit_index // self.num_qubits_per_qumode

    def __iter__(self):
        """Iterate over the list of lists representing the qubits for each qumode in the register"""
        return QumodeIterator(self)

    def __getitem__(self, key):
        """Return a list of QisKit Qubit for each indexed qumode

        Args:
            key (slice or int): index into qumode register

        Raises:
            ValueError: if slice or int not provided

        Returns:
            list: ;ost pf qubits from QuantumRegister representing qumode
        """
        start = None
        stop = self.size
        step = None

        if isinstance(key, slice):
            start_index = key.start if key.start else 0
            stop_index = key.stop if key.stop else self.size
            start = self.num_qubits_per_qumode * start_index
            stop = (
                self.num_qubits_per_qumode * stop_index
            ) + self.num_qubits_per_qumode
            step = (key.step * self.num_qubits_per_qumode) if key.step else None
        elif isinstance(key, int):
            start = self.num_qubits_per_qumode * key
            stop = start + self.num_qubits_per_qumode
        else:
            raise ValueError("Must provide slice or int.")

        return self.qreg[start:stop:step]

    def __len__(self):
        """The length of a QumodeRegister is the number of qumodes (not the num_qumodes * num_qubits_per_qumode)"""
        return self.num_qumodes

    def __contains__(self, qubit):
        """Return true if this QumodeRegister contains the given qubit. This allows callers to use `in` python syntax."""
        return qubit in self.qreg


class QumodeIterator:
    """Iterate over the list of lists representing the qubits for each qumode in the register"""

    def __init__(self, register: QumodeRegister):
        self._index = 0
        self._register = register

    def __iter__(self):
        return self

    def __next__(self):
        if self._index < self._register.num_qumodes:
            next = self._register[self._index]
            self._index += 1

            return next
        else:
            raise StopIteration



import functools
import math
from typing import Sequence
import warnings


import c2qa
import numpy
from qiskit.circuit import Instruction, Qubit
from qiskit_aer.noise.passes.local_noise_pass import LocalNoisePass
from qiskit_aer.noise import kraus_error
from qiskit_aer.noise.noiseerror import NoiseError
from qiskit.utils.units import apply_prefix
import scipy


def calculate_kraus(
    photon_loss_rates: Sequence[float],
    time: float,
    circuit: c2qa.CVCircuit,
    op_qubits: Sequence[int],
    qumode_qubit_indices: Sequence[int],
):
    """
    Calculate Kraus operator given number of photons and photon loss rate over specified time.

    Apply Kraus operator to provided qubit_indices only, tensor product with identity for remaining qubits.

    Following equation 44 from Bosonic Oprations and Measurements, Girvin

    Args:
        photon_loss_rates (Sequence[float]): kappas, the rate of photon loss per second for each qumode
        time (float): current duration of time in seconds
        circuit (CVCircuit): cq2a.CVCircuit with ops for N and a
        op_qubits (Sequence[int]): qubit int indices in the given CVCircuit used by the current instruction
        qumode_qubit_indices (Sequence[int]): qumode int indices in the given CVCircuit to test if qubits from instruction are a part of a qumode

    Returns:
        List of Kraus operators
    """

    # Identity for individual qubit
    qubit_eye = numpy.eye(2)

    operators = []
    kraus_tensor = {}
    loss_rate_index = 0
    for op_qubit in op_qubits:
        if qumode_qubit_indices and op_qubit in qumode_qubit_indices:
            qubit = circuit.qubits[op_qubit]
            qmr_index = circuit.get_qmr_index(qubit)
            qumode_index = circuit.qmregs[qmr_index].get_qumode_index(qubit)
            unique_index = f"{qmr_index}-{qumode_index}"

            cutoff = circuit.get_qmr_cutoff(qmr_index)

            # Tensor Kraus operators, if not already done
            if not kraus_tensor.get(
                unique_index, False
            ):  # FIXME need to tensor per qumode, not qumode register
                kraus_ops = __kraus_operators(
                    photon_loss_rates[loss_rate_index],
                    time,
                    cutoff,
                    circuit.ops.get_a(cutoff),
                    circuit.ops.get_N(cutoff),
                )
                operators = __tensor_operators(operators, kraus_ops)
                kraus_tensor[unique_index] = True
                loss_rate_index += 1
        else:
            # Tensor qubit identity
            operators = __tensor_operators(operators, [qubit_eye])

    return operators


def __tensor_operators(current: list, new: list):
    result = []
    if len(current) > 0:
        for current_op in current:
            for new_op in new:
                result.append(numpy.kron(current_op, new_op))
    else:
        result.extend(new)

    return result


def __kraus_operators(photon_loss_rate: float, time: float, cutoff: int, a, n):
    operators = []
    for photons in range(cutoff + 1):
        kraus = math.sqrt(
            math.pow((1 - math.exp(-1 * photon_loss_rate * time)), photons)
            / math.factorial(photons)
        )
        kraus = kraus * scipy.sparse.linalg.expm(-1 * (photon_loss_rate / 2) * time * n)
        kraus = kraus.dot(a**photons)
        operators.append(kraus.toarray())

    return operators


class PhotonLossNoisePass(LocalNoisePass):
    """Add photon loss noise model to a circuit during transpiler transformation pass."""

    def __init__(
        self,
        photon_loss_rates: Sequence[float],
        circuit: c2qa.CVCircuit,
        instructions: Sequence[str] = None,
        qumodes: Sequence[Qubit] = None,
        time_unit: str = "s",
        dt: float = None,
    ):
        """
        Initialize the Photon Loss noise pass

        Args:
            photon_loss_rate (float): kappa, the rate of photon loss per second
            circuit (CVCircuit): cq2a.CVCircuit with ops for N and a, and cutoff
            instructions (str or list[str]): the instructions error applies to
            qumode (Sequence[Qubit]): qumode qubits error noise pass applies to
            time_unit (string): string photon loss rate unit of time (default "s" for seconds)
            dt (float): optional conversion factor for photon_loss_rate and operation duration to seconds
        """

        self._circuit = circuit
        self._time_unit = time_unit
        self._dt = dt

        if instructions is None:
            self._instructions = None
        elif isinstance(instructions, list):
            self._instructions = instructions
        else:
            self._instructions = [instructions]

        if qumodes is None:
            # Apply photon loss to all qumodes by default
            self._qumodes = self._circuit.qumode_qubits
        elif isinstance(qumodes, list):
            self._qumodes = qumodes
        else:
            self._qumodes = [qumodes]

        self._qumode_qubit_indices = circuit.get_qubit_indices(self._qumodes)

        # Calculate the number of qumodes based on the number of times the QumodeRegister index changes for the given qumode qubits
        qmr_to_num_qubits = {}
        for qubit in self._qumodes:
            qmr_index = circuit.get_qmr_index(qubit)
            if qmr_index not in qmr_to_num_qubits:
                qmr_to_num_qubits[qmr_index] = 0
            qmr_to_num_qubits[qmr_index] += 1

        self._num_qumodes = 0
        for qmr_index, num_qubits in qmr_to_num_qubits.items():
            num_qubits_per_qumode = circuit.get_qmr_num_qubits_per_qumode(qmr_index)
            self._num_qumodes += num_qubits // num_qubits_per_qumode

        if isinstance(photon_loss_rates, list):
            self._photon_loss_rates = photon_loss_rates
        else:
            self._photon_loss_rates = [photon_loss_rates]

        # If only one rate was given for multiple qumodes, apply that rate to all qumodes
        if len(self._photon_loss_rates) == 1 and self._num_qumodes > 1:
            self._photon_loss_rates = self._photon_loss_rates * self._num_qumodes

        # Test that we have the correct number of photon loss rates
        if len(self._photon_loss_rates) != self._num_qumodes:
            raise Exception(
                "List of photon loss rates must have same length as number of qumodes! (i.e., one rate per qumode)"
            )

        # Convert photon loss rate to photons per second
        if self._time_unit == "dt":
            self.photon_loss_rates_sec = [
                rate * self._dt for rate in self._photon_loss_rates
            ]
        else:
            conversion = 1.0 / apply_prefix(1.0, self._time_unit)
            self.photon_loss_rates_sec = [
                rate * conversion for rate in self._photon_loss_rates
            ]

        super().__init__(self._photon_loss_error)

    def _photon_loss_error(self, op: Instruction, qubits: Sequence[int]):
        """Return photon loss error on each operand qubit"""
        error = None

        if self.applies_to_instruction(op, qubits):
            if not op.duration:
                if op.duration is None:
                    warnings.warn(
                        "PhotonLossNoisePass ignores instructions without duration,"
                        " you may need to schedule circuit in advance.",
                        UserWarning,
                    )
                return None

            # Qiskit `delay` gates are always for one qubit, see https://qiskit.org/documentation/stubs/qiskit.circuit.QuantumCircuit.delay.html
            if op.name.startswith("delay"):
                warnings.warn(
                    "Qiskit applies delays as single qubit gates. Qumode PhotonLossNoisePass will not be applied"
                )
                return None

            duration = self.duration_to_sec(op)

            kraus_operators = calculate_kraus(
                self.photon_loss_rates_sec,
                duration,
                self._circuit,
                qubits,
                self._qumode_qubit_indices,
            )

            error = kraus_error(kraus_operators)

        return error

    def applies_to_instruction(self, op: Instruction, qubits: Sequence[int]):
        """Test if this PhotonLossNoisePass applies to the given instruction based on its name and qumodes (qubits)"""
        return (self._instructions is None or op.name in self._instructions) and (
            self._qumode_qubit_indices is None
            or any(x in qubits for x in self._qumode_qubit_indices)
        )

    def duration_to_sec(self, op: Instruction):
        """Return the given Instruction's duration in seconds"""
        if op.unit == "dt":
            if self._dt is None:
                raise NoiseError(
                    "PhotonLossNoisePass cannot apply noise to a 'dt' unit duration without a dt time set."
                )
            duration = op.duration * self._dt
        else:
            duration = apply_prefix(op.duration, op.unit)

        return duration



from copy import copy
from pathlib import Path


from c2qa.circuit import CVCircuit
from c2qa.util import trace_out_qubits, simulate


import matplotlib.pyplot as plt
import matplotlib.ticker as tick
import numpy as np
from numpy import array, zeros, real, meshgrid, exp, pi, conj, sqrt
from qiskit.quantum_info import DensityMatrix, Statevector
from qiskit.result import Result
import qutip
import scipy.stats


def simulate_wigner(
    circuit: CVCircuit,
    xvec: np.ndarray,
    shots: int,
    noise_passes=None,
    conditional: bool = True,
    trace: bool = False,
    g=sqrt(2),
    method: str = "clenshaw",
):
    """Simulate the circuit, optionally partial trace the results, and calculate the Wigner function."""
    states, _, _ = simulate(
        circuit,
        shots=shots,
        noise_passes=noise_passes,
        conditional_state_vector=conditional,
    )

    if states:
        if conditional:
            state = states["0x0"]  # even state
            # state = states["0x1"]  # odd state
        else:
            state = states

        if trace:
            density_matrix = trace_out_qubits(circuit, state)
        else:
            density_matrix = state

        wigner_result = _wigner(density_matrix, xvec, g=g, method=method)
    else:
        print(
            "WARN: No state vector returned by simulation -- unable to calculate Wigner function!"
        )
        wigner_result = None
        state = None

    return wigner_result, state


def simulate_wigner_multiple_statevectors(
    circuit: CVCircuit,
    xvec: np.ndarray,
    shots: int,
    statevector_label: str,
    num_statevectors: int,
    noise_passes=None,
    trace: bool = False,
    g=sqrt(2),
    method: str = "clenshaw",
):
    """Simulate the circuit, optionally partial trace the results, and calculate the Wigner function on each statevector starting with the given label."""
    state, result, _ = simulate(circuit, shots=shots, noise_passes=noise_passes)

    if len(result.results):
        wigner_results = []
        for num in range(num_statevectors):
            state = result.data()[f"{statevector_label}{num}"]
            if trace:
                density_matrix = trace_out_qubits(circuit, state)
            else:
                density_matrix = state

            wigner_results.append(_wigner(density_matrix, xvec, g=g, method=method))
    else:
        print(
            "WARN: No state vector returned by simulation -- unable to calculate Wigner function!"
        )
        wigner_results = None

    return wigner_results


def wigner(
    state,
    axes_min: int = -6,
    axes_max: int = 6,
    axes_steps: int = 200,
    g=sqrt(2),
    method: str = "clenshaw",
):
    """
    Calculate the Wigner function on the given state vector.

    Args:
        state (array-like): state vector to calculate Wigner function
        cutoff (int): cutoff used during simulation
        axes_min (int, optional): Minimum axes plot value. Defaults to -6.
        axes_max (int, optional): Maximum axes plot value. Defaults to 6.
        axes_steps (int, optional): Steps between axes ticks. Defaults to 200.
        hbar (int, optional): hbar value to use in Wigner function calculation. Defaults to 2.

    Returns:
        array-like: Results of Wigner function calculation
    """
    xvec = np.linspace(axes_min, axes_max, axes_steps)
    return _wigner(state, xvec, g=g, method=method)


def wigner_mle(
    states,
    axes_min: int = -6,
    axes_max: int = 6,
    axes_steps: int = 200,
    g=sqrt(2),
    method: str = "clenshaw",
):
    """
    Find the maximum likelihood estimation for the given state vectors and calculate the Wigner function on the result.

    Args:
        states (array-like of array-like): state vectors to calculate MLE and Wigner function
        cutoff (int): cutoff used during simulation
        axes_min (int, optional): Minimum axes plot value. Defaults to -6.
        axes_max (int, optional): Maximum axes plot value. Defaults to 6.
        axes_steps (int, optional): Steps between axes ticks. Defaults to 200.
        hbar (int, optional): hbar value to use in Wigner function calculation. Defaults to 2.

    Returns:
        array-like: Results of Wigner function calculation
    """
    mle_state = []

    # Prevent DeprecationWarning from Qiskit returning Statevector instead of array
    states_data = [state.data for state in states]

    for qubit_states in zip(*states_data):
        # TODO what distribution are the qubit states? (using normal)
        # scipy.stats normal distribution defaults to MLE fit, returns tuple[0] mean, tuple[1] std dev
        mle = scipy.stats.norm.fit(qubit_states)
        mle_state.append(mle[0])

    mle_normalized = mle_state / np.linalg.norm(mle_state)

    return wigner(mle_normalized, axes_min, axes_max, axes_steps, g=g, method=method)


def _wigner(state, xvec, yvec=None, g=sqrt(2), method: str = "clenshaw"):
    if isinstance(state, DensityMatrix):
        rho = state.data
    else:
        rho = DensityMatrix(state).data

    if not yvec:
        yvec = xvec

    return qutip.wigner(psi=qutip.Qobj(rho), xvec=xvec, yvec=yvec, g=g, method=method)


def plot_wigner(
    circuit: CVCircuit,
    state_vector: Statevector,
    trace: bool = True,
    file: str = None,
    axes_min: int = -6,
    axes_max: int = 6,
    axes_steps: int = 200,
    num_colors: int = 100,
    draw_grid: bool = False,
    dpi: int = 100,
    g=sqrt(2),
    method: str = "clenshaw",
):
    """Produce a Matplotlib figure for the Wigner function on the given state vector.

    Optionally perform partial trace.

    Args:
        circuit (CVCircuit): circuit with results to trace (to find Qubit index)
        state_vector (Statevector): simulation results to trace over and plot
        trace (bool, optional): True if qubits should be traced. Defaults to True.
        file (str, optional): File path to save plot. If none, return plot. Defaults to None.
        axes_min (int, optional): Minimum axes plot value. Defaults to -6.
        axes_max (int, optional): Maximum axes plot value. Defaults to 6.
        axes_steps (int, optional): Steps between axes ticks. Defaults to 200.
        num_colors (int, optional): Number of color gradients in legend. Defaults to 100.
        draw_grid (bool, optional): True if grid lines should be drawn on plot. Defaults to False.
    """
    if trace:
        state = trace_out_qubits(circuit, state_vector)
    else:
        state = state_vector

    w_fock = wigner(
        state=state,
        axes_min=axes_min,
        axes_max=axes_max,
        axes_steps=axes_steps,
        g=g,
        method=method,
    )

    plot(
        data=w_fock,
        axes_min=axes_min,
        axes_max=axes_max,
        axes_steps=axes_steps,
        file=file,
        num_colors=num_colors,
        draw_grid=draw_grid,
        dpi=dpi,
    )


def plot(
    data,
    axes_min: int = -6,
    axes_max: int = 6,
    axes_steps: int = 200,
    file: str = None,
    num_colors: int = 100,
    draw_grid: bool = False,
    dpi=100,
):
    """Contour plot the given data array"""
    xvec = np.linspace(axes_min, axes_max, axes_steps)

    amax = np.amax(data)
    amin = np.amin(data)
    if amax == 0 and amin == 0:
        amax = 1
        amin = -1
    abs_max = max(amax, abs(amin))
    color_levels = np.linspace(-abs_max, abs_max, num_colors)

    fig, ax = plt.subplots(constrained_layout=True)
    cont = ax.contourf(xvec, xvec, data, color_levels, cmap="RdBu")

    xvec_int = [int(x) for x in xvec]
    xvec_int = sorted(set(xvec_int))
    ax.set_xlabel(r"$x$")
    # ax.set_xticks(xvec_int)
    ax.set_ylabel(r"$p$")
    # ax.set_yticks(xvec_int)
    ax.set_aspect("equal", "box")
    if draw_grid:
        ax.grid()

    cb = fig.colorbar(cont, ax=ax, format=tick.FormatStrFormatter("%.2f"))
    cb.set_label(r"$W(x,p)$", rotation=270, labelpad=25)

    if file:
        plt.savefig(file, dpi=dpi)
    else:
        plt.show()


def plot_wigner_projection(
    circuit: CVCircuit,
    qubit,
    file: str = None,
    draw_grid: bool = False,
    g=sqrt(2),
    method: str = "clenshaw",
):
    """Plot the projection onto 0, 1, +, - for the given circuit.

    This is limited to CVCircuit with only one qubit, also provided as a parameter.

    Args:
        circuit (CVCircuit): circuit to simulate and plot
        qubit (Qubit): qubit to measure
        file (str, optional): File path to save file, if None return plot. Defaults to None.
        draw_grid (bool, optional): True if gridlines should be drawn on plots. Defaults to False.
    """
    # Get unaltered state vector and partial trace
    x, _, _ = simulate(circuit)
    xT = x.data.conjugate().transpose()

    # Project onto 0 and 1 using Pauli Z
    circuit.z(qubit)
    y, _, _ = simulate(circuit)
    yT = y.data.conjugate().transpose()

    x_xT = x.data * xT
    x_yT = x.data * yT
    y_xT = y.data * xT
    y_yT = y.data * yT

    trace_x_xT = trace_out_qubits(circuit, x_xT)
    trace_x_yT = trace_out_qubits(circuit, x_yT)
    trace_y_xT = trace_out_qubits(circuit, y_xT)
    trace_y_yT = trace_out_qubits(circuit, y_yT)

    projection_zero = (trace_x_xT + trace_x_yT + trace_y_xT + trace_y_yT) / 4
    projection_one = (trace_x_xT - trace_x_yT - trace_y_xT + trace_y_yT) / 4

    # Clean up by popping off the Pauli Z
    circuit.data.pop()

    # Project onto + and - using Pauli X
    circuit.x(qubit)
    y, _, _ = simulate(circuit)
    yT = y.data.conjugate().transpose()

    x_xT = x.data * xT
    x_yT = x.data * yT
    y_xT = y.data * xT
    y_yT = y.data * yT

    trace_x_xT = trace_out_qubits(circuit, x_xT)
    trace_x_yT = trace_out_qubits(circuit, x_yT)
    trace_y_xT = trace_out_qubits(circuit, y_xT)
    trace_y_yT = trace_out_qubits(circuit, y_yT)

    projection_plus = (trace_x_xT + trace_x_yT + trace_y_xT + trace_y_yT) / 4
    projection_minus = (trace_x_xT - trace_x_yT - trace_y_xT + trace_y_yT) / 4

    # Clean up by popping of the Pauli X
    circuit.data.pop()

    # Calculate Wigner functions
    xvec = np.linspace(-6, 6, 200)
    wigner_zero = _wigner(projection_zero, xvec, g=g, method=method)
    wigner_one = _wigner(projection_one, xvec, g=g, method=method)
    wigner_plus = _wigner(projection_plus, xvec, g=g, method=method)
    wigner_minus = _wigner(projection_minus, xvec, g=g, method=method)

    # Plot using matplotlib on four subplots, at double the default width & height
    fig, ((ax0, ax1), (ax2, ax3)) = plt.subplots(2, 2, figsize=(12.8, 12.8))

    _add_contourf(ax0, fig, "Projection onto zero", xvec, xvec, wigner_zero, draw_grid)
    _add_contourf(ax1, fig, "Projection onto one", xvec, xvec, wigner_one, draw_grid)
    _add_contourf(ax2, fig, "Projection onto plus", xvec, xvec, wigner_plus, draw_grid)
    _add_contourf(
        ax3, fig, "Projection onto minus", xvec, xvec, wigner_minus, draw_grid
    )

    # Save to file or display
    if file:
        plt.savefig(file)
    else:
        plt.show()


def plot_wigner_snapshot(
    circuit: CVCircuit,
    result: Result,
    folder: Path = None,
    trace: bool = True,
    axes_min: int = -6,
    axes_max: int = 6,
    axes_steps: int = 200,
    num_colors: int = 100,
    g=sqrt(2),
    method: str = "clenshaw",
):
    for cv_snapshot_id in range(circuit.cv_snapshot_id):
        label = f"cv_snapshot_{cv_snapshot_id}"

        if folder:
            file = Path(folder, f"{label}.png")
        else:
            file = f"{label}.png"

        snapshot = result.data()[label]
        # index = 0
        # if len(snapshot) > 1:
        #     print(f"Simulation had {len(snapshot)} shots, plotting last one")
        #     index = len(snapshot) - 1

        # plot_wigner(circuit, snapshot[index], trace, file, axes_min, axes_max, axes_steps, num_colors, g=g, method=method)
        plot_wigner(
            circuit,
            snapshot,
            trace,
            file,
            axes_min,
            axes_max,
            axes_steps,
            num_colors,
            g=g,
            method=method,
        )


def _add_contourf(ax, fig, title, x, y, z, draw_grid: bool = False):
    """Add a matplotlib contourf plot with color levels based on min/max values in z."""
    amax = np.amax(z)
    amin = abs(np.amin(z))
    max_value = max(amax, amin, 0.0001)  # Force a range if amin/amax are equal
    color_levels = np.linspace(-max_value, max_value, 100)

    cont = ax.contourf(x, y, z, color_levels, cmap="RdBu")

    xvec_int = [int(value) for value in x]
    xvec_int = sorted(set(xvec_int))
    ax.set_xlabel("x")
    ax.set_xticks(xvec_int)

    yvec_int = [int(value) for value in y]
    yvec_int = sorted(set(yvec_int))
    ax.set_ylabel("p")
    ax.set_yticks(yvec_int)

    if draw_grid:
        ax.grid()

    ax.set_title(title)
    fig.colorbar(cont, ax=ax)



import c2qa
import numpy
from pathlib import Path
import qiskit
from qiskit.visualization import plot_histogram
from qiskit.quantum_info import Statevector, DensityMatrix


def test_trace_out_zero(capsys):
    with capsys.disabled():
        qmr = c2qa.QumodeRegister(num_qumodes=1, num_qubits_per_qumode=2)
        qr = qiskit.QuantumRegister(size=1)
        circuit = c2qa.CVCircuit(qmr, qr)

        circuit.initialize([0, 1], qr[0])
        circuit.cv_initialize(0, qmr[0])

        state, result, fock_counts = c2qa.util.simulate(circuit)
        trace = c2qa.util.trace_out_qubits(circuit, state)

        assert state.dims() == (2, 2, 2)
        assert trace.dims() == (2, 2)
        prob = trace.probabilities_dict()
        numpy.testing.assert_almost_equal(prob["00"], 1.0)

        # print("Partial trace Fock state zero")
        # print(DensityMatrix(state).probabilities_dict())
        # print(trace.probabilities_dict())


def test_trace_out_one(capsys):
    with capsys.disabled():
        qmr = c2qa.QumodeRegister(num_qumodes=1, num_qubits_per_qumode=2)
        qr = qiskit.QuantumRegister(size=1)
        circuit = c2qa.CVCircuit(qmr, qr)

        circuit.initialize([1, 0], qr[0])
        circuit.cv_initialize(1, qmr[0])

        state, result, fock_counts = c2qa.util.simulate(circuit)
        trace = c2qa.util.trace_out_qubits(circuit, state)

        assert state.dims() == (2, 2, 2)
        assert trace.dims() == (2, 2)
        prob = trace.probabilities_dict()
        numpy.testing.assert_almost_equal(prob["01"], 1.0)

        # print("Partial trace Fock state one")
        # print(DensityMatrix(state).probabilities_dict())
        # print(trace.probabilities_dict())


def test_trace_out_qubit(capsys):
    with capsys.disabled():
        qmr = c2qa.QumodeRegister(num_qumodes=1, num_qubits_per_qumode=2)
        qbr = qiskit.QuantumRegister(size=1)
        circuit = c2qa.CVCircuit(qmr, qbr)

        circuit.initialize([1, 0], qbr[0])
        circuit.cv_initialize(1, qmr[0])

        state, result, fock_counts = c2qa.util.simulate(circuit)
        trace = c2qa.util.cv_partial_trace(circuit, state, qbr[0])

        assert state.dims() == (2, 2, 2)
        assert trace.dims() == (2, 2)
        prob = trace.probabilities_dict()
        numpy.testing.assert_almost_equal(prob["01"], 1.0)

        # print("Partial trace Fock state one")
        # print(DensityMatrix(state).probabilities_dict())
        # print(trace.probabilities_dict())


def test_trace_out_qumode(capsys):
    with capsys.disabled():
        qmr = c2qa.QumodeRegister(num_qumodes=1, num_qubits_per_qumode=2)
        qbr = qiskit.QuantumRegister(size=1)
        circuit = c2qa.CVCircuit(qmr, qbr)

        circuit.initialize([1, 0], qbr[0])
        circuit.cv_initialize(1, qmr[0])

        state, result, fock_counts = c2qa.util.simulate(circuit)
        trace = c2qa.util.cv_partial_trace(circuit, state, qmr[0])


def test_measure_all_xyz(capsys):
    with capsys.disabled():
        qmr = c2qa.QumodeRegister(num_qumodes=1, num_qubits_per_qumode=4)
        qr = qiskit.QuantumRegister(size=1)
        circuit = c2qa.CVCircuit(qmr, qr)

        # dist = numpy.sqrt(numpy.pi) / numpy.sqrt(2)
        dist = numpy.sqrt(2)

        # qr[0] will init to zero
        circuit.cv_initialize(0, qmr[0])

        circuit.h(qr[0])
        circuit.cv_c_d(dist, qmr[0], qr[0])

        (
            (state_x, result_x, _),
            (state_y, result_y, _),
            (state_z, result_z, _),
        ) = c2qa.util.measure_all_xyz(circuit)

        print("state_x.probabilities_dict()")
        print(state_x.probabilities_dict())

        print("result_x.get_counts() calculated probabilities")
        print(c2qa.util.get_probabilities(result_x))

        print("result_x.to_dict()")
        print(result_x.to_dict())

        plot_histogram(result_x.get_counts(), title="X", figsize=(9, 7)).savefig(
            "tests/plot_histogram_x.png"
        )
        plot_histogram(result_y.get_counts(), title="Y", figsize=(9, 7)).savefig(
            "tests/plot_histogram_y.png"
        )
        plot_histogram(result_z.get_counts(), title="Z", figsize=(9, 7)).savefig(
            "tests/plot_histogram_z.png"
        )


def test_stateread(capsys):
    with capsys.disabled():
        num_qumodes = 2
        qubits_per_mode = 3

        qmr = c2qa.QumodeRegister(
            num_qumodes=num_qumodes, num_qubits_per_qumode=qubits_per_mode, name="qmr"
        )

        circuit = c2qa.CVCircuit(qmr)

        circuit.cv_initialize(2, qmr[0])
        circuit.cv_initialize(0, qmr[1])
        state, result, fock_counts = c2qa.util.simulate(circuit)
        c2qa.util.stateread(
            state,
            numberofqubits=0,
            numberofmodes=num_qumodes,
            cutoff=qmr.cutoff,
            verbose=True,
        )


def test_fockmap(capsys):
    with capsys.disabled():

        # Build rand array of rand dim between 1 and 100, use fockmap to populate initally empty array, and assert that final array is equal to rand array
        for _ in range(10):  # Repeat test 10 times
            dim = numpy.random.randint(low=1, high=101)
            randarray = numpy.random.uniform(low=0, high=1, size=(dim, dim))

            testmatrix = numpy.zeros((dim, dim))
            for i in range(dim):
                for j in range(dim):
                    testmatrix = c2qa.util.fockmap(testmatrix, j, i, randarray[i, j])

            assert (testmatrix == randarray).all()

        # Check fockmap using numpy.outer
        matrix = numpy.zeros((4, 4))
        assert (
            c2qa.util.fockmap(matrix, 0, 0) == numpy.outer([1, 0, 0, 0], [1, 0, 0, 0])
        ).all()  # |0><0|
        assert (
            c2qa.util.fockmap(matrix, 1, [3, 2], [1, 0.5])
            == (
                numpy.outer([0, 0, 0, 1], [0, 1, 0, 0])
                + 0.5 * numpy.outer([0, 0, 1, 0], [0, 1, 0, 0])
            )
        ).all()  # |3><1| + 0.5|2><1|
        assert (
            c2qa.util.fockmap(matrix, 1, [3, 2, 1])
            == c2qa.util.fockmap(matrix, [1, 1, 1], [3, 2, 1])
        ).all()  # |3><1| + |2><1| + |1><1|

        # Check the types which are accepted for each arg.
        for i in range(10):
            # Nested list, numpy.ndarray
            m_types = [[[0, 0], [0, 0]], numpy.zeros((2, 2))]

            # int, list
            fi_types = [0, [1, 0]]

            # int, list
            fo_types = [1, [1, 0]]

            # int, float, complex, empty list, list, numpy.ndarray
            amp_types = [1, 1.0, 1j, [], [1, 1], numpy.array([1, 1])]

            # Generate random indices to test for
            m_index = numpy.random.randint(0, 2)
            fi_index = numpy.random.randint(0, 2)
            fo_index = numpy.random.randint(0, 2)

            if (fi_index == 0) & (fo_index == 0):
                amp_index = numpy.random.randint(0, 4)
            else:
                amp_index = numpy.random.randint(3, 5)

            # Assert that output is a numpy.ndarray
            assert (
                type(
                    c2qa.util.fockmap(
                        m_types[m_index],
                        fi_types[fi_index],
                        fo_types[fo_index],
                        amp_types[amp_index],
                    )
                )
                == numpy.ndarray
            )


def test_circuit_avg_photon_num(capsys):
    with capsys.disabled():
        # Create two qumode registers containing 2 qumodes and 1 qumode respectively.
        qmr1 = c2qa.QumodeRegister(2, 3)
        qmr2 = c2qa.QumodeRegister(1, 3)
        circ = c2qa.CVCircuit(qmr1, qmr2)

        # Initialize the three qumodes to |3>, |4>, |5> Fock states.
        circ.cv_initialize(3, qmr1[0])  # Qumode in |3>
        circ.cv_initialize(4, qmr1[1])  # Qumode in |4>
        circ.cv_initialize(5, qmr2[0])  # Qumode in |5>

        # Print out the indices of qubits in qumodes, grouped by qumode
        print(circ.qumode_qubits_indices_grouped)
        ## >> [[0, 1, 2], [3, 4, 5], [6, 7, 8]]

        # Obtain state
        state, _, _ = c2qa.util.simulate(circ)

        avg_photon_num = c2qa.util.avg_photon_num(circ, state)
        print(avg_photon_num)
        assert [3.0, 4.0, 5.0] == avg_photon_num


def test_qumode_avg_photon_num(capsys):
    with capsys.disabled():
        for _ in range(5):  # Repeat test 5 times
            # Decimals
            decimals = numpy.random.randint(1, 6)

            # Generate random vector
            dim = numpy.random.randint(1, 11)
            vector = numpy.random.uniform(-1, 1, dim) + 1.0j * numpy.random.uniform(
                -1, 1, dim
            )

            # Compute magnitude of each element within vector, and norm of vector
            element_norm = numpy.multiply(vector, numpy.conjugate(vector))
            norm = numpy.sum(element_norm)

            # Dot product between number operator and magnitude
            avg_num = (
                numpy.mean(numpy.dot(element_norm, numpy.array(range(dim)))) / norm
            )

            # Average photon number of statevector, density matrix, and random vector must all match
            assert (
                c2qa.util.qumode_avg_photon_num(Statevector(vector), decimals)
                == c2qa.util.qumode_avg_photon_num(DensityMatrix(vector), decimals)
                == round(avg_num.real, decimals)
            )
